from pydantic import validate_call

from ramanchada2.io.HSDS import read_cha
from ramanchada2.misc.spectrum_deco import add_spectrum_constructor

from ..spectrum import Spectrum


@add_spectrum_constructor(set_applied_processing=False)
@validate_call
def from_chada(filename: str, dataset: str = '/raw', h5module=None):
    x, y, meta = read_cha(filename, dataset, h5module=h5module)
    return Spectrum(x=x, y=y, metadata=meta)  # type: ignore



from __future__ import annotations

import logging
from copy import deepcopy
from typing import Dict, List, Optional, Set, Tuple, Union

import numpy as np
import numpy.typing as npt
from pydantic import PositiveInt, validate_call
from scipy.signal import convolve, savgol_coeffs, savgol_filter
from scipy.stats import median_abs_deviation, rv_histogram

from ramanchada2.io.HSDS import write_cha, write_nexus
from ramanchada2.io.output.write_csv import write_csv as io_write_csv
from ramanchada2.misc.plottable import Plottable
from ramanchada2.misc.types import PositiveOddInt, SpeMetadataModel
from ramanchada2.misc.types.spectrum import (SpeProcessingListModel,
                                             SpeProcessingModel)

logger = logging.getLogger(__name__)


class Spectrum(Plottable):
    _available_processings: Set[str] = set()

    @validate_call(config=dict(arbitrary_types_allowed=True))
    def __init__(self,
                 x: Union[npt.NDArray, int, None] = None,
                 y: Union[npt.NDArray, None] = None,
                 cachefile: Optional[str] = None,
                 metadata: Union[SpeMetadataModel, None] = None,
                 applied_processings: Union[SpeProcessingListModel, None] = None):
        super(Plottable, self).__init__()
        if x is not None:
            if isinstance(x, int):
                self.x = np.arange(x)
            else:
                self.x = x
        if y is not None:
            self.y = y

        self._x_err: Union[npt.NDArray, None] = None
        self._y_err: Union[npt.NDArray, None] = None

        sort_idx = np.argsort(self.x)
        if (np.diff(sort_idx) != 1).any():
            self.x = self.x[sort_idx]
            if self.y is not None:
                self.y = self.y[sort_idx]

        self._cachefile = cachefile
        self._metadata = deepcopy(metadata or SpeMetadataModel(root={}))
        self._applied_processings = deepcopy(applied_processings or SpeProcessingListModel(root=[]))

    def __copy__(self):
        return Spectrum(
            x=self.x,
            y=self.y,
            cachefile=self._cachefile,
            metadata=self._metadata,
            applied_processings=self._applied_processings,
        )

    def __repr__(self):
        return self._applied_processings.repr()

    def applied_processings_dict(self):
        return self._applied_processings.to_list()

    def __str__(self):
        return str(self._applied_processings.to_list())

    def write_csv(self, filename, delimiter=',', newline='\n'):
        csv = io_write_csv(self.x, self.y, delimiter=delimiter)
        with open(filename, 'w', newline=newline) as f:
            for c in csv:
                f.write(c)

    def write_cha(self, chafile, dataset):
        write_cha(chafile, dataset, self.x, self.y, self.meta.serialize())

    def write_nexus(self, chafile, dataset):
        write_nexus(chafile, dataset, self.x, self.y, self.meta.serialize())

    def write_cache(self):
        if self._cachefile:
            self.write_cha(
                self._cachefile,
                '/cache/'+self._applied_processings.cache_path()+'/_data')

    def process(self, algorithm: str, **kwargs):
        if algorithm not in self._available_processings:
            raise ValueError('Unknown algorithm {algorithm}')
        return getattr(self, algorithm)(**kwargs)

    @classmethod
    @validate_call(config=dict(arbitrary_types_allowed=True))
    def apply_creator(cls, step: SpeProcessingModel, cachefile_=None):
        proc = getattr(cls, step.proc)
        spe = proc(*step.args, **step.kwargs, cachefile_=cachefile_)
        return spe

    @validate_call(config=dict(arbitrary_types_allowed=True))
    def apply_processing(self, step: SpeProcessingModel):
        proc = getattr(self, step.proc)
        spe = proc(*step.args, **step.kwargs)
        return spe

    def _plot(self, ax, *args, **kwargs):
        ax.errorbar(
            self.x,
            self.y,
            xerr=self.x_err,
            yerr=self.y_err,
            **kwargs
        )

    def _sort_x(self):
        idx = np.argsort(self.x)
        self.x = self.x[idx]
        self.y = self.y[idx]

    @property
    def x(self): return np.array(self._xdata)

    @x.setter
    def x(self, val: npt.NDArray[np.float64]):
        self._xdata = val
        self._xdata.flags.writeable = False

    @property
    def x_bin_boundaries(self):
        return np.concatenate((
            [(3*self.x[0] - self.x[1])/2],
            (self.x[1:] + self.x[:-1])/2,
            [(3*self.x[-1] - self.x[-2])/2]
        ))

    @property
    def y(self) -> npt.NDArray[np.float64]:
        return np.array(self._ydata)

    @y.setter
    def y(self, val: npt.NDArray[np.float64]):
        self._ydata = val
        self._ydata.flags.writeable = False

    @property
    def y_noise(self):
        return self.y_noise_savgol()

    def y_noise_MAD(self):
        return median_abs_deviation(np.diff(self.y))

    @validate_call(config=dict(validate_default=True))
    def y_noise_savgol_DL(self, order: PositiveOddInt = 1):
        npts = order + 2
        ydata = self.y - np.min(self.y)
        summ = np.sum((ydata - savgol_filter(ydata, npts, order))**2)
        coeff = savgol_coeffs(npts, order)
        coeff[(len(coeff) - 1) // 2] -= 1
        scale = np.sqrt(np.sum(coeff**2))
        return np.sqrt(summ/len(ydata))/scale

    @validate_call(config=dict(validate_default=True))
    def y_noise_savgol(self, order: PositiveOddInt = 1):
        npts = order + 2

        # subtract smoothed signal from original
        coeff = - savgol_coeffs(npts, order)
        coeff[(len(coeff)-1)//2] += 1

        # normalize coefficients so that `sum(coeff**2) == 1`
        coeff /= np.sqrt(np.sum(coeff**2))

        # remove the common floor
        ydata = self.y - np.min(self.y)
        return np.std(convolve(ydata, coeff, mode='same'))

    @property
    def x_err(self):
        if self._x_err is None:
            return np.zeros_like(self._xdata)
        else:
            return self._x_err

    @x_err.setter
    def x_err(self, val: Union[npt.NDArray, None]):
        if val is not None:
            if val.shape != self._xdata.shape:
                raise ValueError(
                    'x_err should have same shape as xdata, expected {self._xdata.shape}, got {val.shape}')
        self._x_err = val

    @property
    def y_err(self):
        if self._y_err is None:
            return np.zeros_like(self._ydata)
        else:
            return self._y_err

    @y_err.setter
    def y_err(self, val: Union[npt.NDArray, None]):
        if val is not None:
            if val.shape != self._ydata.shape:
                raise ValueError(
                    'y_err should have same shape as ydata, expected {self._ydata.shape}, got {val.shape}')
        self._y_err = val

    @property
    def meta(self) -> SpeMetadataModel:
        return self._metadata

    @meta.setter
    @validate_call(config=dict(arbitrary_types_allowed=True))
    def meta(self, val: Union[Dict, SpeMetadataModel]):
        if isinstance(val, dict):
            self._metadata = SpeMetadataModel.model_validate(val)
        else:
            self._metadata = val

    @property
    def result(self):
        return self.meta['ramanchada2_filter_result']

    @result.setter
    def result(self, res: Union[Dict, List]):
        return self.meta._update(dict(ramanchada2_filter_result=res))

    @validate_call(config=dict(arbitrary_types_allowed=True))
    def spe_distribution(self, trim_range: Union[Tuple[float, float], None] = None):
        x_all = self.x_bin_boundaries
        if trim_range is not None:
            l_idx = int(np.argmin(np.abs(x_all - trim_range[0])))
            r_idx = int(np.argmin(np.abs(x_all - trim_range[1])))
            spe_dist = rv_histogram((self.y[l_idx:r_idx], x_all[l_idx:r_idx+1]))
        else:
            spe_dist = rv_histogram((self.y, x_all))
        return spe_dist

    @validate_call(config=dict(arbitrary_types_allowed=True))
    def gen_samples(self, size: PositiveInt, trim_range=None):
        spe_dist = self.spe_distribution(trim_range=trim_range)
        samps = spe_dist.rvs(size=size)
        return samps



import re
from collections import UserList, defaultdict
from typing import List

import numpy as np
import pandas as pd
from lmfit.model import Model, ModelResult, Parameters

from ..plottable import Plottable
from .peak_candidates import ListPeakCandidateMultiModel
from ramanchada2.spectrum.spectrum import Spectrum


class FitPeaksResult(UserList, Plottable):
    def valuesdict(self):
        ret = dict()
        for i in self:
            ret.update(i.params.valuesdict())
        return ret

    @property
    def locations(self):
        return np.array([v for peak in self for k, v in peak.values.items() if k.endswith('center')])

    @property
    def centers(self):
        return np.array([v for peak in self for k, v in peak.values.items() if k.endswith('center')])

    @property
    def fwhm(self):
        return np.array([v for peak in self for k, v in peak.values.items() if k.endswith('fwhm')])

    def boundaries(self, n_sigma=5):
        bounds = list()
        for group in self:
            pos = np.array([v for k, v in group.values.items() if k.endswith('center')])
            sig = np.array([v for k, v in group.values.items() if k.endswith('fwhm')])
            sig /= 2.35
            sig *= n_sigma
            bounds.append((np.min(pos - sig), np.max(pos+sig)))
        return bounds

    def center_amplitude(self, threshold):
        return np.array([
            (v.value, peak.params[k[:-6] + 'amplitude'].value)
            for peak in self
            for k, v in peak.params.items()
            if k.endswith('center')
            if hasattr(v, 'stderr') and v.stderr is not None and v.stderr < threshold
        ]).T

    @property
    def centers_err(self):
        return np.array([
            (v.value, v.stderr)
            for peak in self
            for k, v in peak.params.items()
            if k.endswith('center')
            if hasattr(v, 'stderr') and v.stderr is not None
            ])

    @property
    def fwhms(self):
        return np.array([v for peak in self for k, v in peak.values.items() if k.endswith('fwhm')])

    @property
    def amplitudes(self):
        return np.array([v for peak in self for k, v in peak.values.items() if k.endswith('amplitude')])

    def dumps(self):
        return [peak.dumps() for peak in self]

    @classmethod
    def loads(cls, json_str: List[str]):
        self = cls()
        for p in json_str:
            params = Parameters()
            modres = ModelResult(Model(lambda x: x, None), params)
            self.append(modres.loads(p))
        return self

    def _plot(self, ax, peak_candidate_groups=None, individual_peaks=False, xarr=None,
              label=None,  # ignore label from kwargs
              **kwargs):
        def group_plot(x, fitres, label=''):
            if individual_peaks:
                color = None
                for component in fitres.eval_components(x=x).values():
                    line, = ax.plot(x, component, color=color,
                                    label=(label if color is None else None),
                                    **kwargs)
                    color = line.get_c()
            else:
                ax.plot(x, fitres.eval(x=x), **kwargs)

        if peak_candidate_groups is None:
            for group_i, (bound, fitres) in enumerate(zip(self.boundaries(), self)):
                x = np.linspace(*bound, 200)
                group_plot(x, fitres, label=f'group {group_i}')
        elif isinstance(peak_candidate_groups, ListPeakCandidateMultiModel):
            for group_i, (cand, fitres) in enumerate(zip(peak_candidate_groups, self)):
                x = np.linspace(*cand.boundaries, 2000)
                group_plot(x, fitres, label=f'group {group_i}')
        else:
            for group_i, fitres in enumerate(self):
                left, right = peak_candidate_groups[group_i].boundaries(n_sigma=3)
                x = np.linspace(left, right, 100)
                group_plot(x, fitres, label=f'group {group_i}')

    def to_dataframe(self):
        return pd.DataFrame(
            [
                dict(name=f'g{group:02d}_{key}', value=val.value, stderr=val.stderr)
                for group, res in enumerate(self)
                for key, val in res.params.items()
            ]
        ).sort_values('name')

    def to_dataframe_peaks(self):
        regex = re.compile(r'p([0-9]+)_(.*)')
        ret = defaultdict(dict)
        for group_i, group in enumerate(self):
            for par in group.params:
                m = regex.match(par)
                if m is None:
                    continue
                peak_i, par_name = m.groups()
                ret[f'g{group_i:02d}_p{peak_i}'][par_name] = group.params[par].value
                ret[f'g{group_i:02d}_p{peak_i}'][f'{par_name}_stderr'] = group.params[par].stderr
        return pd.DataFrame.from_dict(ret, orient='index')

    def to_csv(self, path_or_buf=None, sep=',', **kwargs):
        return self.to_dataframe_peaks().to_csv(path_or_buf=path_or_buf, sep=sep, **kwargs)

    def gen_fake_spectrum(self, xarr):
        summ = np.zeros_like(xarr)
        last_i = 0
        last_y = 0
        for m in self:
            mx = m.userkws['x']
            xi = np.searchsorted(xarr, mx)
            evalm = m.eval()
            summ[xi] = evalm
            summ[np.arange(last_i, xi[0])] = np.interp(np.arange(last_i, xi[0]), [last_i, xi[0]], [last_y, evalm[0]])
            last_i = xi[-1]
            last_y = evalm[-1]
        fake_spe = Spectrum(x=xarr, y=summ)
        return fake_spe



#!/usr/bin/env python3

from __future__ import annotations

import numpy as np

from ramanchada2.misc.base_class import BaseClass
from ramanchada2.misc.plottable import Plottable
from .spectral_peak import SpectralPeak


class SpectralComponentCollection(Plottable, BaseClass):
    def __init__(self, peaks, **kwargs):
        super(Plottable, self).__init__()
        super(BaseClass, self).__init__()
        self._peaks = set(peaks)
        self.kwargs = kwargs
        self.reset_origin()

    def reset_origin(self):
        self._origin = [(type(self).__name__,
                         [tuple(sorted(self._peaks, key=lambda x: repr(x)))],
                         self.kwargs)]

    def __call__(self, x):
        ret = np.array([p(x) for p in self._peaks]).sum(axis=0)
        return ret

    def get_deltas(self):
        pos, ampl = zip(*[p.delta for p in self._peaks])
        return pos, ampl

    def get_curve(self):
        ...

    @property
    def limit_3sigma(self):
        lims = [p.limit_3sigma for p in self._peaks]
        return np.min(lims), np.max(lims)

    def __iadd__(self, peak: SpectralPeak):
        self._peaks.add(peak)
        self.reset_origin()

    def _plot(self, ax, draw='combined line', **kwargs):
        if draw == 'deltas':
            stem_kwargs = dict(basefmt='', markerfmt='rD')
            stem_kwargs.update(kwargs)
            ax.stem(*self.get_deltas(), **stem_kwargs)
        elif draw == 'crosses':
            x0, a, fwhm = zip(*[i.pos_amp_fwhm for i in self._peaks])
            x0 = np.array(x0)
            a = np.array(a)
            fwhm = np.array(fwhm)
            err_kwargs = dict(linewidth=0, elinewidth=1)
            err_kwargs.update(kwargs)
            ax.errorbar(x0, a/2, xerr=fwhm/2, yerr=a/2, **err_kwargs)
        elif draw == 'combined line':
            x = np.arange(*self.limit_3sigma)
            ax.plot(x, self(x), **kwargs)
        elif draw == 'individual lines':
            for p in self._peaks:
                p.plot(ax, draw='line', **kwargs)
        else:
            raise TypeError("draw can be 'combined line', 'individual lines', 'crosses' or 'deltas'")



"""Create spectrum from simulation output files."""

from io import TextIOBase
from typing import Literal, Union

import numpy as np
from pydantic import validate_call

from ramanchada2.io.simulated import read_simulated_lines
from ramanchada2.misc.spectrum_deco import add_spectrum_constructor

from ..spectrum import Spectrum


@add_spectrum_constructor()
@validate_call(config=dict(arbitrary_types_allowed=True))
def from_simulation(in_file: Union[str, TextIOBase],
                    sim_type: Literal['vasp', 'crystal_out', 'crystal_dat', 'raw_dat'],
                    use: Literal['I_tot', 'I_perp', 'I_par',
                                 'I_xx', 'I_xy', 'I_xz',
                                 'I_yy', 'I_yz', 'I_zz'] = 'I_tot'
                    ):
    """
    Generate spectrum from simulation file.

    The returned spectrum has only few x/y pairs -- one for each simulated line. Values along
    the x-axis will not be uniform. To make it uniform, one needs to resample the spectrum.

    Args:
        in_file:
            Path to a local file, or file-like object.
        sim_type:
            If `vasp`: `.dat` file from VASP simulation. If `crystal_out`: `.out` file from CRYSTAL simulation, not
            preferred. If `crystal_dat`: `.dat` file from CRYSTAL simulation.
        use:
            `I_tot`, `I_perp`, `I_par`, `I_xx`, `I_xy`, `I_xz`, `I_yy`, `I_yz`, `I_zz`, `I_tot`, `I_perp`, `I_par` are
            available for both CRYSTAL and VASP. `I_xx`, `I_xy`, `I_xz`, `I_yy`, `I_yz`, `I_zz` are available only for
            CRYSTAL.
    """
    if isinstance(in_file, TextIOBase):
        labels, x, ydict = read_simulated_lines(in_file, sim_type=sim_type, use={use})
    else:
        with open(in_file) as f:
            labels, x, ydict = read_simulated_lines(f, sim_type=sim_type, use={use})
    y = ydict[use]
    spe = Spectrum(x=np.array(x), y=np.array(y))
    spe._sort_x()
    return spe



from pydantic import validate_call

from ramanchada2.misc.spectrum_deco import (add_spectrum_filter,
                                            add_spectrum_method)
from ramanchada2.misc.utils.ramanshift_to_wavelength import \
    abs_nm_to_shift_cm_1 as util_abs_nm_to_shift_cm_1
from ramanchada2.misc.utils.ramanshift_to_wavelength import \
    shift_cm_1_to_abs_nm as util_shift_cm_1_to_abs_nm

from ..spectrum import Spectrum


@add_spectrum_method
@validate_call(config=dict(arbitrary_types_allowed=True))
def abs_nm_to_shift_cm_1(spe: Spectrum, /,
                         laser_wave_length_nm: float):
    return util_abs_nm_to_shift_cm_1(spe.x, laser_wave_length_nm=laser_wave_length_nm)


@add_spectrum_method
@validate_call(config=dict(arbitrary_types_allowed=True))
def shift_cm_1_to_abs_nm(spe: Spectrum, /,
                         laser_wave_length_nm: float):
    return util_shift_cm_1_to_abs_nm(spe.x, laser_wave_length_nm=laser_wave_length_nm)


@add_spectrum_filter
@validate_call(config=dict(arbitrary_types_allowed=True))
def abs_nm_to_shift_cm_1_filter(old_spe: Spectrum,
                                new_spe: Spectrum, /,
                                laser_wave_length_nm: float):
    new_spe.x = util_abs_nm_to_shift_cm_1(old_spe.x, laser_wave_length_nm=laser_wave_length_nm)


@add_spectrum_filter
@validate_call(config=dict(arbitrary_types_allowed=True))
def shift_cm_1_to_abs_nm_filter(old_spe: Spectrum,
                                new_spe: Spectrum, /,
                                laser_wave_length_nm: float):
    new_spe.x = util_shift_cm_1_to_abs_nm(old_spe.x, laser_wave_length_nm=laser_wave_length_nm)



import numpy as np
from pydantic import PositiveInt, validate_call
from scipy import signal

from ramanchada2.misc.spectrum_deco import add_spectrum_filter

from ..spectrum import Spectrum


@add_spectrum_filter
@validate_call(config=dict(arbitrary_types_allowed=True))
def moving_average(old_spe: Spectrum,
                   new_spe: Spectrum, /,
                   window_size: PositiveInt = 10):
    """
    Moving average filter.

    Args:
        window_size:
            `int`, optional, default is `10`.
    """
    y = [np.average(old_spe.y[i:min(i + window_size, len(old_spe.y))])
         for i in range(len(old_spe.y))]
    new_spe.y = np.array(y)


@add_spectrum_filter
@validate_call(config=dict(arbitrary_types_allowed=True))
def moving_average_convolve(old_spe: Spectrum,
                            new_spe: Spectrum, /,
                            window_size: PositiveInt = 10):
    new_spe.y = signal.convolve(old_spe.y, np.ones(window_size)/window_size, mode='same')



import logging
from typing import Optional

from pydantic import validate_call

import ramanchada2.misc.types.spectrum as spe_t
from ramanchada2.misc.spectrum_deco import add_spectrum_constructor

from ..spectrum import Spectrum

logger = logging.getLogger(__name__)


@add_spectrum_constructor(set_applied_processing=False)
@validate_call
def from_cache_or_calc(required_steps: spe_t.SpeProcessingListModel,
                       cachefile: Optional[str] = None):
    def recall():
        if len(required_steps):
            last_proc = required_steps.pop()
            if last_proc.is_constructor:
                spe = Spectrum.apply_creator(last_proc, cachefile_=cachefile)
            else:
                spe = recur(required_steps=required_steps)
                spe._cachefile = cachefile
                spe = spe.apply_processing(last_proc)
            return spe
        else:
            raise Exception('no starting point')

    def recur(required_steps: spe_t.SpeProcessingListModel):
        try:
            if cachefile:
                spe = get_cache()
            else:
                spe = recall()
        except Exception:
            spe = recall()
        spe._cachefile = cachefile
        return spe

    def get_cache():
        try:
            cache_path = required_steps.cache_path()
            if cache_path:
                cache_path = '/cache/'+cache_path+'/_data'
            else:
                cache_path = 'raw'
            spe = Spectrum.from_chada(cachefile, cache_path)
            spe._applied_processings.extend_left(required_steps.root)
            return spe
        except Exception as e:
            logger.info(repr(e))
            raise e

    return recur(required_steps)



from typing import Union

import numpy as np
from numpy.typing import NDArray
from pydantic import validate_call

from ramanchada2.misc.spectrum_deco import add_spectrum_filter

from ..spectrum import Spectrum


@add_spectrum_filter
@validate_call(config=dict(arbitrary_types_allowed=True))
def __add__(
        old_spe: Spectrum,
        new_spe: Spectrum,
        arg: Union[Spectrum, NDArray, float]):
    if isinstance(arg, Spectrum):
        if not (old_spe.x == arg.x).all():
            ValueError('x axes should be equal')
        new_spe.y = old_spe.y + arg.y
    elif isinstance(arg, np.ndarray):
        if old_spe.y.shape != arg.shape:
            ValueError(f'shapes does not match {old_spe.y.shape} != {arg.shape}')
        new_spe.y = old_spe.y + arg
    elif isinstance(arg, float):
        new_spe.y = old_spe.y + arg
    else:
        ValueError('This should never happen')



from typing import Dict, TextIO, Tuple

import numpy as np
from numpy.typing import NDArray

from .bw_format import bw_format
from .neegala_format import neegala_format
from .rruf_format import rruf_format

""" There are 4 types of TXT data files that can be distinguished by their first line:
    1. File Version;BWRam4.11_11
    2. File Version;BWSpec4.11_1
    3. <wavenumber>	<intensity>
    4. #Wave		#Intensity
      <wavenumber>	<intensity>
    """


def read_txt(data_in: TextIO) -> Tuple[NDArray, NDArray, Dict]:
    lines = data_in.readlines()
    if lines[0].startswith('File Version;BW'):
        data, meta = bw_format(lines)
        positions = data['RamanShift'].to_numpy()
        intensities = data['DarkSubtracted#1'].to_numpy()
        meta['@axes'] = ['RamanShift']
        meta['@signal'] = 'DarkSubtracted'
    elif lines[0].startswith('##'):
        # rruf format
        positions, intensities, meta = rruf_format(lines)
        meta['@axes'] = ['']
        meta['@signal'] = ''
    elif ',' in lines[0] and not lines[0].split(',')[0].isdigit():
        data, meta = neegala_format(lines)
        positions = data['Raman_Shift'].to_numpy()
        intensities = data['Processed_Data'].to_numpy()
        meta['@axes'] = ['Raman_Shift']
        meta['@signal'] = 'Processed_Data'
    else:  # assume two column spectrum
        meta = dict()
        if lines[0].startswith('#'):
            # assume header row
            data = np.genfromtxt(lines, names=True, loose=False)
            meta['@axes'] = [data.dtype.names[0]]
            meta['@signal'] = data.dtype.names[1]
            data = np.array(data.tolist())
        else:
            data = np.genfromtxt(lines, loose=False)
            meta['@axes'] = ['']
            meta['@signal'] = ''
        positions, intensities = data.T
    return positions, intensities, meta



import re
from io import TextIOBase
from typing import List

import pandas as pd
from pydantic import validate_call

from ramanchada2.misc.exceptions import InputParserError


@validate_call(config=dict(arbitrary_types_allowed=True))
def lines_from_crystal_out(data_in: TextIOBase) -> pd.DataFrame:
    def advance_to(content: str) -> None:
        while content not in data_in.readline():
            continue

    def read_paragraph() -> List[str]:
        ret = list()
        while True:
            line = data_in.readline().rstrip()
            if line == '':
                break
            ret.append(line)
        return ret

    def parse(regex, lines: List[str]) -> List[List[str]]:
        ret = list()
        if len(lines) < 2:
            raise InputParserError()
        for match in [regex.match(line) for line in lines[2:]]:
            if match:
                ret.append(match.groups())
            else:
                raise InputParserError()
        return ret

    def skip_line():
        data_in.readline()

    advance_to('POLYCRYSTALLINE ISOTROPIC INTENSITIES (ARBITRARY UNITS)')
    skip_line()  # empty line
    polyXtal_lines = read_paragraph()
    advance_to('SINGLE CRYSTAL DIRECTIONAL INTENSITIES (ARBITRARY UNITS)')
    skip_line()  # empty line
    monoXtal_lines = read_paragraph()
    # data_in is processed

    polyXtal_regex = re.compile(r'\s*(\d+)-\s*(\d+)\s*([\d.]+)\s*\((\w+)\s*\)' + r'\s*([\d.]+)'*3)
    polyXtal_parsed = parse(polyXtal_regex, polyXtal_lines)
    polyXtal_df = pd.DataFrame.from_records(
        polyXtal_parsed,
        columns=['ModeL', 'ModeU', 'Frequencies', 'Origin', 'I_tot', 'I_par', 'I_perp'])
    polyXtal_df = polyXtal_df.astype(
        dict(zip(polyXtal_df.keys(), [*[int]*2, float, str, *[float]*3])))

    monoXtal_regex = re.compile(r'\s*(\d+)-\s*(\d+)\s*([\d.]+)\s*\((\w+)\s*\)' + r'\s*([\d.]+)'*6)
    monoXtal_parsed = parse(monoXtal_regex, monoXtal_lines)
    monoXtal_df = pd.DataFrame.from_records(
        monoXtal_parsed,
        columns=['ModeL', 'ModeU', 'Frequencies', 'Origin', 'I_xx', 'I_xy', 'I_xz', 'I_yy', 'I_yz', 'I_zz'])
    monoXtal_df = monoXtal_df.astype(
        dict(zip(monoXtal_df.keys(), [*[int]*2, float, str, *[float]*6])))
    merge = pd.merge(polyXtal_df, monoXtal_df)
    return merge



#!/usr/bin/env python3

from __future__ import annotations
from abc import ABC, abstractmethod

import numpy as np
from uncertainties import unumpy

from ramanchada2.spectral_components.spectral_component import SpectralComponent


class SpectralPeak(SpectralComponent, ABC):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _plot(self, ax, draw='line', **kwargs):
        if draw == 'line':
            x = np.arange(*self.limit_3sigma)
            y = self(x)
            ax.errorbar(unumpy.nominal_values(x), unumpy.nominal_values(y),
                        yerr=unumpy.std_devs(y), **kwargs)
        elif draw == 'delta':
            args = dict(basefmt='', markerfmt='rD')
            args.update(kwargs)
            ax.stem(*self.delta, **args)
        elif draw == 'cross':
            x0, a, fwhm = self.pos_amp_fwhm
            ax.errorbar(x0, a/2, xerr=fwhm/2, yerr=a/2, **kwargs)
        else:
            raise TypeError("draw can be 'line', 'cross' or 'delta'")

    @abstractmethod
    def delta(self): pass

    @abstractmethod
    def limit_3sigma(self): pass

    @abstractmethod
    def pos_amp_fwhm(self): pass



import numpy as np


def shift_cm_1_to_abs_nm_dict(deltas, laser_wave_length_nm):
    arr = np.array(list(deltas.items()), dtype=float)
    arr[:, 0] = 1/(1/laser_wave_length_nm - arr[:, 0] * 1e-7)
    return dict(arr)


def abs_nm_to_shift_cm_1_dict(deltas, laser_wave_length_nm):
    arr = np.array(list(deltas.items()), dtype=float)
    arr[:, 0] = 1e7*(1/laser_wave_length_nm - 1/arr[:, 0])
    return dict(arr)


def abs_nm_to_shift_cm_1(wl, laser_wave_length_nm):
    return 1e7*(1/laser_wave_length_nm - 1/wl)


def shift_cm_1_to_abs_nm(wn, laser_wave_length_nm):
    shift_nm = wn * 1e-7
    absolute_nm = 1/(1/laser_wave_length_nm - shift_nm)
    return absolute_nm


def laser_wl_nm(raman_shift_cm_1, wave_length_nm):
    return 1/(1/wave_length_nm+raman_shift_cm_1*1e-7)



#!/usr/bin/env python3

from io import TextIOBase

import pandas as pd
from pydantic import validate_call


@validate_call(config=dict(arbitrary_types_allowed=True))
def lines_from_crystal_dat(data_in: TextIOBase) -> pd.DataFrame:
    return pd.DataFrame(
        data=[
            lin.split()
            for lin in data_in.readlines()
            if not lin.startswith('#')
        ],
        columns=[
            'Frequencies', 'I_tot', 'I_par', 'I_perp',
            'I_xx', 'I_xy', 'I_xz', 'I_yy', 'I_yz', 'I_zz'
        ],
        dtype=float)



#!/usr/bin/env python3

from .peak_candidates import (ListPeakCandidateMultiModel,
                              PeakCandidateMultiModel, PeakModel)
from .positive_not_multiple import PositiveOddInt
from .spectrum import *  # noqa



#!/usr/bin/env python3

from typing import TextIO, Tuple, Dict

from numpy.typing import NDArray
import numpy as np


def read_csv(data_in: TextIO) -> Tuple[NDArray, NDArray, Dict]:
    lines = data_in.readlines()
    positions, intensities = np.genfromtxt(lines, delimiter=',', dtype=float).T
    filter_nan = ~(
        np.isnan(positions) |
        np.isnan(intensities)
    )
    positions = positions[filter_nan]
    intensities = intensities[filter_nan]
    meta: Dict[str, None] = dict()
    return positions, intensities, meta



import numpy as np
from pydantic import validate_call

from ramanchada2.misc.spectrum_deco import add_spectrum_filter

from ..spectrum import Spectrum


@add_spectrum_filter
@validate_call(config=dict(arbitrary_types_allowed=True))
def add_poisson_noise(
        old_spe: Spectrum,
        new_spe: Spectrum, /,
        scale: float = 1,
        # validation for rng_seed is removed because
        # it makes in-place modification impossible
        rng_seed=None):
    r"""
    Add poisson noise to the spectrum.
    For each particular sample the noise is proportional to $\sqrt{scale*a_i}$.

    Args:
        scale:
            `float`, optional, default is `1`. Scale the amplitude of the noise.
        rng_seed:
            `int` or rng state, optional. Seed for the random generator. If a state is provided, it is updated
            in-place.
    """
    if isinstance(rng_seed, dict):
        rng = np.random.default_rng()
        rng.bit_generator.state = rng_seed
    else:
        rng = np.random.default_rng(rng_seed)
    dat = old_spe.y + [rng.normal(0., np.sqrt(i*scale)) for i in old_spe.y]
    dat[dat < 0] = 0
    if isinstance(rng_seed, dict):
        rng_seed.update(rng.bit_generator.state)
    new_spe.y = np.array(dat)



from pydantic import PositiveInt

# For more info check https://github.com/pydantic/pydantic/issues/10111
# PositiveOddInt = typing.Annotated[int,
#                                   annotated_types.Ge(0),
#                                   annotated_types.Not(annotated_types.MultipleOf(2))]

# FIXME This is a quickfix
PositiveOddInt = PositiveInt



from typing import Literal, Tuple

import numpy as np
from pydantic import validate_call

from ramanchada2.misc.spectrum_deco import add_spectrum_filter

from ..spectrum import Spectrum


@add_spectrum_filter
@validate_call(config=dict(arbitrary_types_allowed=True))
def trim_axes(old_spe: Spectrum,
              new_spe: Spectrum, /,
              method: Literal['x-axis', 'bins'],
              boundaries: Tuple[float, float],
              ):
    if method == 'bins':
        lb = int(boundaries[0])
        rb = int(boundaries[1])
    elif method == 'x-axis':
        lb = int(np.argmin(np.abs(old_spe.x - boundaries[0])))
        rb = int(np.argmin(np.abs(old_spe.x - boundaries[1])))
    new_spe.x = old_spe.x[lb:rb]
    new_spe.y = old_spe.y[lb:rb]



import numpy.typing as npt
from pydantic import validate_call

from ramanchada2.misc.spectrum_deco import add_spectrum_filter

from ..spectrum import Spectrum


@add_spectrum_filter
@validate_call(config=dict(arbitrary_types_allowed=True))
def set_new_xaxis(old_spe: Spectrum,
                  new_spe: Spectrum, /,
                  xaxis: npt.NDArray):
    if old_spe.x.shape != xaxis.shape:
        raise ValueError('Shape of xaxis should match the shape of xaxis of the spectrum')
    new_spe.x = xaxis



import numpy as np
from pydantic import PositiveFloat, validate_call

from ramanchada2.misc.spectrum_deco import add_spectrum_filter

from ..spectrum import Spectrum


@add_spectrum_filter
@validate_call(config=dict(arbitrary_types_allowed=True))
def drop_spikes(old_spe: Spectrum,
                new_spe: Spectrum, /,
                n_sigma: PositiveFloat = 10):
    """
    Removes single-bin spikes.

    Args:
        n_sigma:
            `float`, optional, default is `10`.
    """
    yi = old_spe.y[1:-1]
    yi_1 = old_spe.y[:-2]
    yi1 = old_spe.y[2:]
    y_merit = np.abs(2*yi-yi_1-yi1) - np.abs(yi1-yi_1)
    use_idx = y_merit < n_sigma * y_merit.std()
    use_idx = np.concatenate(([True], use_idx, [True]))
    new_spe.x = old_spe.x[use_idx]
    new_spe.y = old_spe.y[use_idx]


@add_spectrum_filter
@validate_call(config=dict(arbitrary_types_allowed=True))
def recover_spikes(old_spe: Spectrum,
                   new_spe: Spectrum, /,
                   n_sigma: PositiveFloat = 10):
    """
    Recover single-bin spikes.

    Args:
        n_sigma:
            `float`, optional, default is `10`.
    """
    yi = old_spe.y[1:-1]
    yi_1 = old_spe.y[:-2]
    yi1 = old_spe.y[2:]
    y_merit = np.abs(2*yi-yi_1-yi1) - np.abs(yi1-yi_1)
    use_idx = y_merit < n_sigma * y_merit.std()
    use_idx = np.concatenate(([True], use_idx, [True]))
    new_spe.y = np.interp(old_spe.x, old_spe.x[use_idx], old_spe.y[use_idx])


@add_spectrum_filter
@validate_call(config=dict(arbitrary_types_allowed=True))
def get_spikes(old_spe: Spectrum,
               new_spe: Spectrum, /,
               n_sigma: PositiveFloat = 10):
    """
    Get single-bin spikes only.

    Args:
        n_sigma:
            `float`, optional, default is `10`.
    """
    yi = old_spe.y[1:-1]
    yi_1 = old_spe.y[:-2]
    yi1 = old_spe.y[2:]
    y_merit = np.abs(2*yi-yi_1-yi1) - np.abs(yi1-yi_1)
    use_idx = y_merit < n_sigma * y_merit.std()
    use_idx = np.concatenate(([True], use_idx, [True]))
    new_spe.x = old_spe.x[~use_idx]
    new_spe.y = old_spe.y[~use_idx]



import numpy as np
from pydantic import PositiveFloat, validate_call

from ramanchada2.misc.spectrum_deco import add_spectrum_filter

from ..spectrum import Spectrum


@add_spectrum_filter
@validate_call(config=dict(arbitrary_types_allowed=True))
def add_gaussian_noise(
        old_spe: Spectrum,
        new_spe: Spectrum, /,
        sigma: PositiveFloat,
        # validation for rng_seed is removed because
        # it makes in-place modification impossible
        rng_seed=None):
    r"""
    Add gaussian noise to the spectrum.
    Random number i.i.d. $N(0, \sigma)$ is added to every sample

    Args:
        sigma:
            Sigma of the gaussian noise.
        rng_seed:
            `int` or rng state, optional, seed for the random generator. If a state is provided, it is updated
            in-place.
    """
    if isinstance(rng_seed, dict):
        rng = np.random.default_rng()
        rng.bit_generator.state = rng_seed
    else:
        rng = np.random.default_rng(rng_seed)
    dat = old_spe.y + rng.normal(0., sigma, size=len(old_spe.y))
    if any(dat < 0):
        dat += abs(dat.min())
    if isinstance(rng_seed, dict):
        rng_seed.update(rng.bit_generator.state)
    new_spe.y = np.array(dat)



from io import TextIOBase

import pandas as pd
from pydantic import validate_call


@validate_call(config=dict(arbitrary_types_allowed=True))
def lines_from_raw_dat(data_in: TextIOBase) -> pd.DataFrame:
    df = pd.read_table(data_in, sep=r'\s+', dtype=float)
    df.rename(columns={'#FREQUENCIES': 'Frequencies'}, inplace=True)
    return df



import logging
from typing import Dict, Tuple

import h5py
import numpy as np
import numpy.typing as npt
from pydantic import validate_call

from ramanchada2.misc.exceptions import ChadaReadNotFoundError

logger = logging.getLogger()


# https://manual.nexusformat.org/examples/napi/python.html
# https://manual.nexusformat.org/examples/python/simple_example_basic/index.html
@validate_call(config=dict(arbitrary_types_allowed=True))
def write_nexus(filename: str,
                dataset: str,
                x: npt.NDArray, y: npt.NDArray, meta: Dict, h5module=None):
    _h5 = h5module or h5py
    try:
        with _h5.File(filename, 'a') as f:
            f.attrs['default'] = dataset
            try:
                nxentry = f.require_group('sample')
            except:  # noqa: E722
                pass

            nxentry = f.require_group('instrument')
            for m in meta:
                print(m, meta[m])

            try:
                nxentry = f.require_group(dataset)
                nxentry.attrs["NX_class"] = 'NXentry'
                nxentry.attrs['default'] = 'data'
            except:  # noqa: E722
                pass

            try:
                nxdata = nxentry.require_group('data')
                nxdata.attrs["NX_class"] = 'NXdata'
                nxdata.attrs['signal'] = 'spectrum'
                nxdata.attrs['axes'] = 'raman_shift'
                nxdata.attrs['raman_shift_indices'] = [0,]
            except:  # noqa: E722
                pass

            try:
                tth = nxdata.require_group('raman_shift', data=x)
                tth.attrs['units'] = 'cm-1'
                tth.attrs['long_name'] = 'Raman shift (cm-1)'
            except:  # noqa: E722
                pass

            try:
                counts = nxdata.create_dataset('spectrum', data=y)
                counts.attrs['units'] = 'au'
                counts.attrs['long_name'] = 'spectrum'
            except:  # noqa: E722
                pass

    except ValueError as e:
        logger.warning(repr(e))


class DatasetExistsError(Exception):
    pass


@validate_call(config=dict(arbitrary_types_allowed=True))
def write_cha(filename: str,
              dataset: str,
              x: npt.NDArray, y: npt.NDArray, meta: Dict, h5module=None):
    data = np.stack([x, y])
    try:
        _h5 = h5module or h5py
        with _h5.File(filename, mode='a') as h5:
            if h5.get(dataset) is None:
                ds = h5.create_dataset(dataset, data=data)
                ds.attrs.update(meta)
            else:
                raise DatasetExistsError(f'dataset `{dataset}` already exists in file `{filename}`')
    except ValueError as e:
        raise e


def read_cha(filename: str,
             dataset: str, h5module=None
             ) -> Tuple[npt.NDArray, npt.NDArray, Dict]:
    _h5 = h5module or h5py
    with _h5.File(filename, mode='r') as h5:
        data = h5.get(dataset)
        if data is None:
            raise ChadaReadNotFoundError(f'dataset `{dataset}` not found in file `{filename}`')
        x, y = data[:]
        meta = dict(data.attrs)
    return x, y, meta


def filter_dataset(topdomain, domain, process_file, sample=None, wavelength=None, instrument=None,
                   provider=None, investigation=None, kwargs={}, h5module=None):
    _h5 = h5module or h5py
    with _h5.File(domain) as dataset:
        if (sample is not None) and (dataset["annotation_sample"].attrs["sample"] == sample):
            process_file(topdomain, domain, **kwargs)


def visit_domain(topdomain="/", process_dataset=None, kwargs={}, h5module=None):
    _h5 = h5module or h5py
    if topdomain.endswith("/"):
        with _h5.Folder(topdomain) as domain:
            domain._getSubdomains()
            for domain in domain._subdomains:
                if domain["class"] == "folder":
                    visit_domain("{}/".format(domain["name"]), process_dataset, kwargs, h5module=_h5)
                else:
                    if not (process_dataset is None):
                        process_dataset(topdomain, domain["name"], **kwargs, h5module=_h5)
    else:
        if not (process_dataset is None):
            process_dataset(None, topdomain, **kwargs, h5module=_h5)



from typing import Dict, List, Tuple

import pandas
from numpy.typing import NDArray


def rruf_format(lines: List[str]) -> Tuple[NDArray, NDArray, Dict]:
    for i, ll in enumerate(lines):
        if not ll.startswith('##'):
            start_spe = i
            break
    meta = dict([ll.strip()[2:].split('=') for ll in lines[:start_spe]])
    for i, ll in enumerate(lines):
        if ll.startswith('##END'):
            stop_spe = i
            break
    data = pandas.DataFrame.from_records(
        data=[ll.split(',') for ll in lines[start_spe:stop_spe]]
        ).apply(pandas.to_numeric).dropna(axis=0)
    positions, intensities = data.to_numpy().T
    return positions, intensities, meta



from abc import ABC, abstractmethod

from pydantic import BaseModel, RootModel


class PydBaseModel(BaseModel, ABC):
    model_config = dict(
        arbitrary_types_allowed=True
    )

    @abstractmethod
    def serialize(self):
        pass


class PydRootModel(RootModel, ABC):
    model_config = dict(
        arbitrary_types_allowed=True
    )

    @abstractmethod
    def serialize(self):
        pass



#!/usr/bin/env python3

from functools import wraps
from ramanchada2.spectrum.spectrum import Spectrum
from .dynamically_added import dynamically_added_constructors


class add_spectrum_constructor:
    def __init__(self, set_applied_processing=True):
        self.set_proc = set_applied_processing

    def __call__(self, fun):
        @wraps(fun)
        def retf(*args, cachefile_=None, **kwargs) -> Spectrum:
            spe = fun(*args, **kwargs)
            if self.set_proc:
                spe._applied_processings.assign(proc=fun.__name__, args=args, kwargs=kwargs)
                if cachefile_:
                    spe._cachefile = cachefile_
                spe.write_cache()
            return spe
        if hasattr(Spectrum, fun.__name__):
            raise ValueError(f'redefining {fun.__name__}')
        setattr(Spectrum, fun.__name__, retf)
        dynamically_added_constructors.add(fun.__name__)
        return retf



"""Create spectrum from local files."""

import os
from typing import Literal, Union

import spc_io
from pydantic import validate_call

from ramanchada2.io.experimental import rc1_parser, read_csv, read_txt
from ramanchada2.misc.spectrum_deco import add_spectrum_constructor
from ramanchada2.misc.types import SpeMetadataModel

from ..spectrum import Spectrum


@add_spectrum_constructor()
@validate_call(config=dict(arbitrary_types_allowed=True))
def from_local_file(
        in_file_name: str,
        filetype: Union[None, Literal['spc', 'sp', 'spa', '0', '1', '2',
                                      'wdf', 'ngs', 'jdx', 'dx',
                                      'txt', 'txtr', 'csv', 'prn', 'rruf']] = None,
        backend: Union[None, Literal['native', 'rc1_parser']] = None):
    """
    Read experimental spectrum from a local file.

    Args:
        in_file_name:
            Path to a local file containing a spectrum.
        filetype:
            Specify the filetype. Filetype can be any of: `spc`, `sp`, `spa`, `0`, `1`, `2`, `wdf`, `ngs`, `jdx`, `dx`,
            `txt`, `txtr`, `csv`, `prn`, `rruf` or `None`. `None` used to determine by extension of the file.
        backend:
            `native`, `rc1_parser` or `None`. `None` means both.

    Raises:
        ValueError:
            When called with unsupported file formats.
    """
    def load_native():
        if filetype is None:
            ft = os.path.splitext(in_file_name)[1][1:]
        else:
            ft = filetype
        if ft in {'txt', 'txtr', 'prn', 'rruf'}:
            with open(in_file_name) as fp:
                x, y, meta = read_txt(fp)
        elif ft in {'csv'}:
            with open(in_file_name) as fp:
                x, y, meta = read_csv(fp)
        elif ft in {'spc'}:
            with open(in_file_name, 'rb') as fp:
                spc = spc_io.SPC.from_bytes_io(fp)
                if len(spc) != 1:
                    raise ValueError(f'Single subfile SPCs are supported. {len(spc)} subfiles found')
                x = spc[0].xarray
                y = spc[0].yarray
                meta = spc.log_book.text
        else:
            raise ValueError(f'filetype {ft} not supported')
        meta["Original file"] = os.path.basename(in_file_name)
        spe = Spectrum(x=x, y=y, metadata=meta)  # type: ignore
        return spe

    def load_rc1():
        x, y, meta = rc1_parser.parse(in_file_name, filetype)
        spe = Spectrum(x=x, y=y, metadata=SpeMetadataModel.model_validate(meta))
        return spe

    if backend == 'native':
        spe = load_native()
    elif backend == 'rc1_parser':
        spe = load_rc1()
    elif backend is None:
        try:
            spe = load_native()
        except Exception:
            spe = load_rc1()
    spe._sort_x()
    return spe



from typing import Literal

import numpy as np
import statsmodels.api as sm
from pydantic import validate_call
from scipy.ndimage import gaussian_filter1d
from scipy.signal import medfilt, savgol_filter, wiener
from scipy.signal.windows import boxcar

from ramanchada2.misc.spectrum_deco import add_spectrum_filter

from ..spectrum import Spectrum


@add_spectrum_filter
@validate_call(config=dict(arbitrary_types_allowed=True))
def smoothing_RC1(old_spe: Spectrum,
                  new_spe: Spectrum, /, *args,
                  method: Literal['savgol', 'sg',
                                  'wiener',
                                  'median',
                                  'gauss', 'gaussian',
                                  'lowess',
                                  'boxcar'],
                  **kwargs):
    if method == 'savgol' or method == 'sg':
        new_spe.y = savgol_filter(old_spe.y, **kwargs)  # window_length, polyorder
    elif method == 'wiener':
        new_spe.y = wiener(old_spe.y, **kwargs)
    elif method == 'gaussian' or method == 'gauss':
        new_spe.y = gaussian_filter1d(old_spe.y, **kwargs)  # sigma
    elif method == 'median':
        new_spe.y = medfilt(old_spe.y, **kwargs)
    elif method == 'lowess':
        kw = dict(span=11)
        kw.update(kwargs)
        x = np.linspace(0, 1, len(old_spe.y))
        new_spe.y = sm.nonparametric.lowess(old_spe.y, x, frac=(5*kw['span'] / len(old_spe.y)), return_sorted=False)
    elif method == 'boxcar':
        kw = dict(box_pts=11)
        kw.update(kwargs)
        box = boxcar(**kwargs, sym=True)
        new_spe.y = np.convolve(old_spe.y, box, mode='same')



from typing import Any, Callable, Dict, Literal, Optional, Tuple, Union

import numpy as np
from pydantic import PositiveInt, validate_call
from scipy import fft, signal
from scipy.interpolate import (Akima1DInterpolator, CubicSpline,
                               PchipInterpolator)

from ramanchada2.misc.spectrum_deco import (add_spectrum_filter,
                                            add_spectrum_method)

from ..spectrum import Spectrum


@add_spectrum_method
@validate_call(config=dict(arbitrary_types_allowed=True))
def resample_NUDFT(spe: Spectrum, /,
                   x_range: Tuple[float, float] = (0, 4000),
                   xnew_bins: PositiveInt = 100,
                   window: Optional[Union[Callable,
                                          Tuple[Any, ...],  # e.g. ('gaussian', sigma)
                                          Literal['barthann', 'bartlett', 'blackman', 'blackmanharris',
                                                  'bohman', 'boxcar', 'chebwin', 'cosine', 'dpss',
                                                  'exponential', 'flattop', 'gaussian', 'general_cosine',
                                                  'general_gaussian', 'general_hamming', 'hamming', 'hann',
                                                  'kaiser', 'nuttall', 'parzen', 'taylor', 'triang', 'tukey']
                                          ]] = None,
                   cumulative: bool = False):

    x_new = np.linspace(x_range[0], x_range[1], xnew_bins, endpoint=False)
    x = spe.x
    y = spe.y
    x = np.array(x)
    x_range = (np.min(x_range), np.max(x_range))
    y = y[(x >= x_range[0]) & (x < x_range[1])]
    x = x[(x >= x_range[0]) & (x < x_range[1])]

    w = (x-x_range[0])/(x_range[1]-x_range[0])*np.pi*2
    x -= x_range[0]

    k = np.arange(xnew_bins)

    Y_new = np.sum([yi*np.exp(-1j*wi*k) for yi, wi in zip(y, w)], axis=0)

    if window is None:
        window = 'blackmanharris'

    if hasattr(window, '__call__'):
        h = (window(len(Y_new)*2))[len(Y_new):]  # type: ignore
    else:
        h = signal.windows.get_window(window, len(Y_new)*2)[len(Y_new):]
    Y_new *= h

    y_new = fft.irfft(Y_new, n=xnew_bins)
    if cumulative:
        y_new = np.cumsum(y_new)
        y_new /= y_new[-1]
    return x_new, y_new


@add_spectrum_filter
@validate_call(config=dict(arbitrary_types_allowed=True))
def resample_NUDFT_filter(old_spe: Spectrum,
                          new_spe: Spectrum, /,
                          x_range: Tuple[float, float] = (0, 4000),
                          xnew_bins: PositiveInt = 100,
                          window=None,
                          cumulative: bool = False):
    new_spe.x, new_spe.y = resample_NUDFT(old_spe,
                                          x_range=x_range,
                                          xnew_bins=xnew_bins,
                                          window=window,
                                          cumulative=cumulative)


@add_spectrum_method
@validate_call(config=dict(arbitrary_types_allowed=True))
def resample_spline(spe: Spectrum, /,
                    x_range: Tuple[float, float] = (0, 4000),
                    xnew_bins: PositiveInt = 100,
                    spline: Literal['pchip', 'akima', 'makima', 'cubic_spline'] = 'pchip',
                    interp_kw_args: Optional[Dict] = None,
                    cumulative: bool = False):

    kw_args: Dict[str, Any] = {}
    if spline == 'pchip':
        spline_fn = PchipInterpolator
    elif spline == 'akima':
        spline_fn = Akima1DInterpolator
    elif spline == 'makima':
        spline_fn = Akima1DInterpolator
        kw_args['method'] = 'makima'
    elif spline == 'cubic_spline':
        spline_fn = CubicSpline
        kw_args['bc_type'] = 'natural'

    if interp_kw_args is not None:
        kw_args.update(interp_kw_args)

    x_new = np.linspace(x_range[0], x_range[1], xnew_bins, endpoint=False)
    y_new = spline_fn(spe.x, spe.y, **kw_args)(x_new)

    y_new[np.isnan(y_new)] = 0
    if cumulative:
        y_new = np.cumsum(y_new)
        y_new /= y_new[-1]

    return x_new, y_new


@add_spectrum_filter
@validate_call(config=dict(arbitrary_types_allowed=True))
def resample_spline_filter(old_spe: Spectrum,
                           new_spe: Spectrum, /,
                           x_range: Tuple[float, float] = (0, 4000),
                           xnew_bins: PositiveInt = 100,
                           spline: Literal['pchip', 'akima', 'makima', 'cubic_spline'] = 'pchip',
                           interp_kw_args: Optional[Dict] = None,
                           cumulative: bool = False):
    new_spe.x, new_spe.y = resample_spline(old_spe,
                                           x_range=x_range,
                                           xnew_bins=xnew_bins,
                                           spline=spline,
                                           interp_kw_args=interp_kw_args,
                                           cumulative=cumulative)



class BaseClass(object):
    def __init__(self):
        self._origin = []

    def __repr__(self):
        return '/'.join(
            [i.replace('/', '_') for i in self.origin_list_str]
            ).replace(' ', '')

    @property
    def origin(self):
        return self._origin

    @origin.setter
    def origin(self, value):
        self._origin = list(value)

    @property
    def origin_list_str(self):
        ret = list()
        for name, args, kwargs in self.origin:
            str_args = [str(i) for i in args]
            str_kwargs = [f'{k}={repr(v)}' for k, v in kwargs.items()]
            sss = f'{name}({", ".join(str_args + str_kwargs)})'
            ret.append(sss)
        return ret



import numpy as np
from pydantic import PositiveFloat, confloat, validate_call

from ramanchada2.misc.spectrum_deco import add_spectrum_filter

from ..spectrum import Spectrum


@validate_call(config=dict(arbitrary_types_allowed=True))
def generate_add_gaussian_noise_drift(y, /,
                                      sigma: PositiveFloat,
                                      coef: confloat(ge=0, le=1),  # type: ignore [valid-type]
                                      # validation for rng_seed is removed because
                                      # it makes in-place modification impossible
                                      rng_seed=None):
    if isinstance(rng_seed, dict):
        rng = np.random.default_rng()
        rng.bit_generator.state = rng_seed
    else:
        rng = np.random.default_rng(rng_seed)
    gaus = rng.normal(0., sigma+coef/np.sqrt(2), size=len(y))
    cs = np.cumsum(gaus)
    # coef*sum(cs[:i]) + (1-coef)*gaus is identical to
    # coef*sum(cs[:i-1]) + gaus
    noise = coef*cs + gaus*(1-coef)
    noise -= np.std(noise)
    dat = y + noise
    if any(dat < 0):
        dat += abs(dat.min())
    if isinstance(rng_seed, dict):
        rng_seed.update(rng.bit_generator.state)
    return np.array(dat)


@add_spectrum_filter
@validate_call(config=dict(arbitrary_types_allowed=True))
def add_gaussian_noise_drift(
        old_spe: Spectrum,
        new_spe: Spectrum, /,
        sigma: PositiveFloat,
        coef: confloat(ge=0, le=1),  # type: ignore [valid-type]
        # validation for rng_seed is removed because
        # it makes in-place modification impossible
        rng_seed=None):
    r"""
    Add cumulative gaussian noise to the spectrum.
    Exponential-moving-average-like gaussian noise is added
    to each sample. The goal is to mimic the low-frequency noise
    (or random substructures in spectra).
    The additive noise is
    .. math::
        a_i = coef*\sum_{j=0}^{i-1}g_j + g_i,
    where
    .. math::
        g_i = \mathcal{N}(0, 1+\frac{coef}{\sqrt 2}).
    This way drifting is possible while keeping the
    .. math::
        \sigma(\Delta(a)) \approx 1.

    Args:
        sigma:
            Sigma of the gaussian noise.
        coef:
            `float` in `[0, 1]`, drifting coefficient. If `coef == 0`, the result is identical to
            `add_gaussian_noise()`.
        rng_seed:
            `int` or rng state, optional. Seed for the random generator. If a state is provided, it is updated
            in-place.
    """
    new_spe.y = generate_add_gaussian_noise_drift(old_spe.y,
                                                  sigma=sigma,
                                                  coef=coef,
                                                  rng_seed=rng_seed)



from typing import Dict, List, Literal, Union

import lmfit
import numpy as np
import numpy.typing as npt
from pydantic import NonNegativeInt, validate_call
from scipy import interpolate

from ramanchada2.misc.spectrum_deco import (add_spectrum_filter,
                                            add_spectrum_method)

from ...misc import utils as rc2utils
from ..spectrum import Spectrum


class DeltaSpeModel:
    def __init__(self, deltas: Dict[float, float], shift=0, scale=1):
        components = list()
        self.params = lmfit.Parameters()
        self.minx = np.min(list(deltas.keys()))
        self.maxx = np.max(list(deltas.keys()))
        self.params.add('shift', value=shift, vary=True)
        self.params.add('scale', value=scale, vary=True, min=.1, max=10)
        self.params.add('scale2', value=0, vary=False, min=-.1, max=.1)
        self.params.add('scale3', value=0, vary=False, min=-1e-3, max=1e-3)
        self.params.add('sigma', value=1, vary=True)
        self.params.add('gain', value=1, vary=True)
        for comp_i, (k, v) in enumerate(deltas.items()):
            prefix = f'comp_{comp_i}'
            components.append(lmfit.models.GaussianModel(prefix=f'comp_{comp_i}_'))
            self.params.add(prefix + '_center', expr=f'shift + {k}*scale + {k}**2*scale2 + {k}**3*scale3', vary=False)
            self.params.add(prefix + '_amplitude', expr=f'{v}*gain', vary=False)
            self.params.add(prefix + '_sigma', expr='sigma', vary=False)
        self.model = np.sum(components)

    def fit(self, spe, sigma, ax=None, no_fit=False):
        self.params['sigma'].set(value=sigma if sigma > 1 else 1)
        spe_conv = spe.convolve('gaussian', sigma=sigma)
        if no_fit:
            fit_res = self.model.fit(spe_conv.y, x=spe_conv.x, params=self.params, max_nfev=-1)
        else:
            fit_res = self.model.fit(spe_conv.y, x=spe_conv.x, params=self.params)
        self.params = fit_res.params
        if ax is not None:
            spe_conv.plot(ax=ax)
            ax.plot(spe_conv.x, fit_res.eval(x=spe_conv.x), 'r')


@add_spectrum_method
@validate_call(config=dict(arbitrary_types_allowed=True))
def calibrate_by_deltas_model(spe: Spectrum, /,
                              deltas: Dict[float, float],
                              convolution_steps: Union[None, List[float]] = [15, 1],
                              scale2=True, scale3=False,
                              init_guess: Literal[None, 'cumulative'] = None,
                              ax=None, **kwargs
                              ):
    """
    - Builds a composite model based on a set of user specified delta lines.
    - Initial guess is calculated based on 10-th and 90-th percentiles of
      the distributions.

    The phasespace of the model is flat with big amount of narrow minima.
    In order to find the best fit, the experimental data are successively
    convolved with gaussians with different widths startign from wide to
    narrow. The model for the calibration is 3-th order polynomial, which
    potentialy can be changed for higher order polynomial. In order to avoid
    solving the inverse of the calibration function, the result is tabulated
    and interpolated linarly for each bin of the spectrum.
    This alogrithm is useful for corse calibration.
    """
    mod = DeltaSpeModel(deltas)
    spe_padded = spe

    if init_guess == 'cumulative':
        deltasx = np.array(list(deltas.keys()))

        deltas_cs = np.cumsum(list(deltas.values()))
        deltas_cs /= deltas_cs[-1]

        deltas_idx10 = np.argmin(np.abs(deltas_cs-.1))
        deltas_idx90 = np.argmin(np.abs(deltas_cs-.9))
        x1, x2 = deltasx[[deltas_idx10, deltas_idx90]]

        spe_cs = np.cumsum(
            spe_padded.moving_average(50).subtract_moving_minimum(10).moving_average(5).y)  # type: ignore

        spe_cs /= spe_cs[-1]
        spe_idx10 = np.argmin(np.abs(spe_cs-.1))
        spe_idx90 = np.argmin(np.abs(spe_cs-.9))
        y1, y2 = spe_padded.x[[spe_idx10, spe_idx90]]

        scale = (y1-y2)/(x1-x2)
        shift = -scale * x1 + y1
    else:
        scale = 1
        shift = 0
    gain = np.sum(spe.y)/np.sum(list(deltas.values()))
    mod.params['scale'].set(value=scale)
    mod.params['shift'].set(value=shift)
    mod.params['gain'].set(value=gain)
    mod.params['sigma'].set(value=2.5)

    if ax is not None:
        spe_padded.plot(ax=ax)

    if convolution_steps is not None:
        for sig in convolution_steps:
            mod.fit(spe=spe_padded, sigma=sig, ax=ax, **kwargs)

    if scale2:
        mod.params['scale2'].set(vary=True, value=0)
        # mod.fit(spe_padded, sigma=1, ax=ax, **kwargs)
        mod.fit(spe_padded, sigma=0, ax=ax, **kwargs)
    if scale3:
        mod.params['scale2'].set(vary=True, value=0)
        mod.params['scale3'].set(vary=True, value=0)
        # mod.fit(spe_padded, sigma=1, ax=ax, **kwargs)
        mod.fit(spe_padded, sigma=0, ax=ax, **kwargs)
    return mod.model, mod.params


@add_spectrum_filter
@validate_call(config=dict(arbitrary_types_allowed=True))
def calibrate_by_deltas_filter(old_spe: Spectrum,
                               new_spe: Spectrum, /,
                               deltas: Dict[float, float],
                               convolution_steps,
                               init_guess=None,
                               **kwargs
                               ):
    mod, par = old_spe.calibrate_by_deltas_model(  # type: ignore
        deltas=deltas,
        convolution_steps=convolution_steps,
        init_guess=init_guess,
        **kwargs)

    deltasx = np.array(list(deltas.keys()))
    dxl, dxr = deltasx[[0, -1]]
    xl = dxl - (dxr - dxl)
    xr = dxl + (dxr - dxl)
    true_x = np.linspace(xl, xr, len(old_spe.x)*6)
    meas_x = (par['shift'].value + true_x * par['scale'] +
              true_x**2 * par['scale2'] + true_x**3 * par['scale3'])
    x_cal = np.zeros_like(old_spe.x)
    for i in range(len(old_spe.x)):
        idx = np.argmax(meas_x > old_spe.x[i])
        pt_rto = (old_spe.x[i] - meas_x[idx-1])/(meas_x[idx] - meas_x[idx-1])
        x_cal[i] = (true_x[idx] - true_x[idx-1])*pt_rto + true_x[idx-1]
    new_spe.x = x_cal


@add_spectrum_filter
@validate_call(config=dict(arbitrary_types_allowed=True))
def xcal_fine(old_spe: Spectrum,
              new_spe: Spectrum, /, *,
              ref: Union[Dict[float, float], List[float]],
              should_fit=False,
              poly_order: NonNegativeInt,
              find_peaks_kw={},
              ):

    if isinstance(ref, dict):
        ref_pos = np.array(list(ref.keys()))
    else:
        ref_pos = np.array(ref)

    if should_fit:
        spe_pos_dict = old_spe.fit_peak_positions(center_err_threshold=1, find_peaks_kw=find_peaks_kw)  # type: ignore
    else:
        find_kw = dict(sharpening=None)
        find_kw.update(find_peaks_kw)
        spe_pos_dict = old_spe.find_peak_multipeak(**find_kw).get_pos_ampl_dict()  # type: ignore
    spe_cent = np.array(list(spe_pos_dict.keys()))

    if poly_order == 0:
        p = rc2utils.align_shift(spe_cent, ref_pos)
        spe_cal = old_spe.scale_xaxis_fun(lambda x: x + p)  # type: ignore
    else:
        def cal_func(x, *a):
            return [par*(x/1000)**power for power, par in enumerate(a)]

        p0 = np.resize([0, 1000, 0], poly_order + 1)
        p = rc2utils.align(spe_cent, ref_pos, p0=p0, func=cal_func)
        spe_cal = old_spe.scale_xaxis_fun(  # type: ignore
            (lambda x, *args: np.sum(cal_func(x, *args), axis=0)), args=p)
    new_spe.x = spe_cal.x


@add_spectrum_filter
@validate_call(config=dict(arbitrary_types_allowed=True))
def xcal_fine_RBF(old_spe: Spectrum,
                  new_spe: Spectrum, /, *,
                  ref: Union[Dict[float, float], List[float], npt.NDArray],
                  should_fit=False,
                  kernel: Literal['thin_plate_spline', 'cubic', 'quintic',
                                  'multiquadric', 'inverse_multiquadric',
                                  'inverse_quadratic', 'gaussian',
                                  ] = 'thin_plate_spline',
                  find_peaks_kw={},
                  **kwargs,
                  ):
    """Wavelength calibration using Radial basis fuction interpolation

    Please be cautious! Interpolation might not be the most appropriate
    approach for this type of calibration.

    **kwargs are passed to RBFInterpolator
    """

    if isinstance(ref, dict):
        ref_pos = np.array(list(ref.keys()))
    else:
        ref_pos = np.array(ref)

    if should_fit:
        spe_pos_dict = old_spe.fit_peak_positions(center_err_threshold=1, find_peaks_kw=find_peaks_kw)  # type: ignore
    else:
        find_kw = dict(sharpening=None)
        find_kw.update(find_peaks_kw)
        spe_pos_dict = old_spe.find_peak_multipeak(**find_kw).get_pos_ampl_dict()  # type: ignore
    spe_cent = np.array(list(spe_pos_dict.keys()))

    spe_idx, ref_idx = rc2utils.find_closest_pairs_idx(spe_cent, ref_pos)
    if len(ref_idx) == 1:
        _offset = (ref_pos[ref_idx][0] - spe_cent[spe_idx][0])
        new_spe.x = old_spe.x + _offset
    else:
        kwargs["kernel"] = kernel
        interp = interpolate.RBFInterpolator(spe_cent[spe_idx].reshape(-1, 1), ref_pos[ref_idx], **kwargs)
        new_spe.x = interp(old_spe.x.reshape(-1, 1))



from io import TextIOBase
from typing import Dict, List, Literal, Set, Tuple

from pydantic import validate_call

from .crystal.discrete_lines_dat import lines_from_crystal_dat
from .crystal.discrete_lines_out import lines_from_crystal_out
from .lines_from_raw_dat import lines_from_raw_dat
from .vasp.vasp_simulation_dat import lines_from_vasp_dat


@validate_call(config=dict(arbitrary_types_allowed=True))
def read_simulated_lines(data_in: TextIOBase,
                         sim_type: Literal['vasp', 'crystal_out', 'crystal_dat', 'raw_dat'],
                         use: Set[Literal[
                             'I_tot', 'I_perp', 'I_par',
                             'I_xx', 'I_xy', 'I_xz', 'I_yy', 'I_yz', 'I_zz'
                             ]] = {'I_tot'}
                         ) -> Tuple[List[str], List[float], Dict[str, List[float]]]:
    positions: List[float]
    intensities: Dict[str, List[float]] = dict()
    if sim_type.startswith('crystal'):
        if sim_type.endswith('out'):
            tbl = lines_from_crystal_out(data_in)
        elif sim_type.endswith('dat'):
            tbl = lines_from_crystal_dat(data_in)
        else:
            raise Exception('This should never happen')
        names = [f'_{i}' for i in range(len(tbl['I_tot']))]
        for key in use:
            intensities.update({key: tbl[key].to_list()})
        positions = tbl['Frequencies'].to_list()

    elif sim_type == 'vasp':
        tbl = lines_from_vasp_dat(data_in)
        names = [f'_{mode}'.replace('.', '_') for mode in tbl['mode']]
        for key in use:
            if key in {'I_xx', 'I_xy', 'I_xz', 'I_yy', 'I_yz', 'I_zz'}:
                raise ValueError('vasp simulation does not support monocrystal intensities')
            if key == 'I_tot':
                intensities.update({key: tbl['activity'].to_list()})
            else:
                intensities.update({key: tbl[key].to_list()})
        positions = tbl['freq(cm-1)'].to_list()

    elif sim_type == 'raw_dat':
        tbl = lines_from_raw_dat(data_in)
        names = [''] * len(tbl['I_tot'])
        for key in use:
            intensities.update({key: tbl[key].to_list()})
        positions = tbl['Frequencies'].to_list()
    else:
        raise Exception('This should never happen')
    return names, positions, intensities



from typing import Dict, List, Tuple

import pandas


def neegala_format(lines: List[str]) -> Tuple[pandas.DataFrame, Dict]:
    for i, ll in enumerate(lines):
        if ll.startswith('Pixels,Wavelength,Wavenumbers,Raman_Shift,Raw_Data,Background_Data,Processed_Data'):
            start_spe = i
            break
    else:
        raise ValueError('The input is not neegala format')
    meta = dict([i.split(',', 1) for i in lines[:start_spe]])
    spe_lines = [ll.strip().split(',') for ll in lines[start_spe:]]

    data = pandas.DataFrame.from_records(
        data=spe_lines[1:], columns=spe_lines[0]
        ).apply(pandas.to_numeric).dropna(axis=0)
    return data, meta



from typing import Literal
import numpy as np
from pydantic import validate_call

from ramanchada2.misc.spectrum_deco import add_spectrum_filter
from ..spectrum import Spectrum


@add_spectrum_filter
@validate_call(config=dict(arbitrary_types_allowed=True))
def normalize(old_spe: Spectrum,
              new_spe: Spectrum, /,
              strategy: Literal['unity', 'min_unity', 'unity_density', 'unity_area', 'minmax',
                                'L1', 'L2'] = 'minmax'):
    """
    Normalize the spectrum.

    Args:
        strategy:
            If `unity`: normalize to `sum(y)`. If `min_unity`: subtract the minimum and normalize to 'unity'. If
            `unity_density`: normalize to `Σ(y_i*Δx_i)`. If `unity_area`: same as `unity_density`. If `minmax`: scale
            amplitudes in range `[0, 1]`. If 'L1' or 'L2': L1 or L2 norm without subtracting the pedestal.
    """
    if strategy == 'unity':
        res = old_spe.y
        res /= np.sum(res)
        new_spe.y = res
    elif strategy == 'min_unity':
        res = old_spe.y - np.min(old_spe.y)
        res /= np.sum(res)
        new_spe.y = res
    if strategy == 'unity_density' or strategy == 'unity_area':
        res = old_spe.y
        res /= np.sum(res * np.diff(old_spe.x_bin_boundaries))
        new_spe.y = res
    elif strategy == 'minmax':
        res = old_spe.y - np.min(old_spe.y)
        res /= np.max(res)
        new_spe.y = res
    elif strategy == 'L1':
        res /= np.linalg.norm(res, 1)
        new_spe.y = res
    elif strategy == 'L2':
        res /= np.linalg.norm(res)
        new_spe.y = res



#!/usr/bin/env python3

from typing import Dict, Callable, Union

import numpy as np
from pydantic import validate_call, PositiveInt

from ..spectrum import Spectrum
from ramanchada2.misc.spectrum_deco import add_spectrum_constructor


@add_spectrum_constructor()
@validate_call(config=dict(arbitrary_types_allowed=True))
def from_delta_lines(
        deltas: Dict[float, float],
        xcal: Union[Callable[[float], float], None] = None,
        nbins: PositiveInt = 2000,
        **kwargs
        ):
    """
    Generate `Spectrum` with delta lines.

    Args:
        deltas:
            Keys of the dictionary are the `x` positions of the deltas; values are the amplitudes of the corresponding
            deltas.
        xcal:
            Callable, optional. `x` axis calibration function.
        nbins:
            `int`, optional. Number of bins in the spectrum.

    Example:

    This will produce spectrum with 1000 bins in the range `[-1000, 2000)`:
    ```py
    xcal = lambda x: x*3 -1000, nbins=1000
    ```
    """
    if xcal is None:
        dk = list(deltas.keys())
        dkmin, dkmax = np.min(dk), np.max(dk)
        if dkmin == dkmax:
            dkmin, dkmax = dkmin*.8, dkmax*1.2
        else:
            dkmin -= (dkmax-dkmin) * .1
            dkmax += (dkmax-dkmin) * .1
        x = np.linspace(dkmin, dkmax, nbins, endpoint=False, dtype=float)
    else:
        x = np.linspace(xcal(0), xcal(nbins), nbins, endpoint=False)
    y = np.zeros_like(x)
    for pos, ampl in deltas.items():
        idx = np.argmin(np.abs(x - pos))
        y[idx] += ampl
    spe = Spectrum(x=x, y=y, **kwargs)
    return spe



from typing import List, Union

import numpy as np
import numpy.typing as npt
import pandas as pd
from pydantic import PositiveInt, validate_call
from scipy import linalg
from sklearn.cluster import KMeans
from sklearn.metrics.pairwise import euclidean_distances


def argmin2d(A):
    ymin_idx = np.argmin(A, axis=0)
    xmin_idx = np.argmin(A, axis=1)
    x_idx = np.unique(xmin_idx[xmin_idx[ymin_idx[xmin_idx]] == xmin_idx])
    y_idx = np.unique(ymin_idx[ymin_idx[xmin_idx[ymin_idx]] == ymin_idx])
    matches = np.stack([y_idx, x_idx]).T
    return matches


def find_closest_pairs_idx(x, y):
    outer_dif = np.abs(np.subtract.outer(x, y))
    return argmin2d(outer_dif).T


def find_closest_pairs(x, y):
    x_idx, y_idx = find_closest_pairs_idx(x, y)
    return x[x_idx], y[y_idx]


@validate_call(config=dict(arbitrary_types_allowed=True))
def align(x, y,
          p0: Union[List[float], npt.NDArray] = [0, 1, 0, 0],
          func=lambda x, a0, a1, a2, a3: (a0*np.ones_like(x), a1*x, a2*x**2/1, a3*(x/1000)**3),
          max_iter: PositiveInt = 1000):
    """
    Iteratively finds best match between x and y and evaluates the x scaling parameters.
    min((lambda(x, *p)-y)**2 | *p)
    Finds best parameters *p that minimise L2 distance between scaled x and original y
    """
    if isinstance(p0, list):
        p = np.array(p0)
    else:
        p = p0
    loss = np.inf
    cur_x = x
    for it in range(max_iter):
        cur_x = np.sum(func(x, *p), axis=0)
        x_idx, y_idx = find_closest_pairs_idx(cur_x, y)
        x_match, y_match = x[x_idx], y[y_idx]
        p_bak = p
        obj_mat = np.stack(func(x_match, *np.ones_like(p)), axis=1)
        p, *_ = linalg.lstsq(obj_mat, y_match, cond=1e-8)
        loss_bak = loss
        loss = np.sum((x_match-y_match)**2)/len(x_match)**2
        if np.allclose(p, p_bak):
            break
        if loss > loss_bak:
            pass
            return p_bak
    return p


@validate_call(config=dict(arbitrary_types_allowed=True))
def align_shift(x, y,
                p0: float = 0,
                max_iter: PositiveInt = 1000):
    loss = np.inf
    cur_x = x
    p = p0
    for it in range(max_iter):
        cur_x = x + p
        x_idx, y_idx = find_closest_pairs_idx(cur_x, y)
        x_match, y_match = x[x_idx], y[y_idx]
        p_bak = p
        p = np.mean(y_match-x_match)
        loss_bak = loss
        loss = np.sum((y_match-x_match)**2)
        if np.allclose(p, p_bak):
            break
        if loss > loss_bak:
            return p_bak
    return p


def match_peaks(spe_pos_dict, ref):
    # Min-Max normalize the reference values
    min_value = min(ref.values())
    max_value = max(ref.values())
    if len(ref.keys()) > 1:
        normalized_ref = {key: (value - min_value) / (max_value - min_value) for key, value in ref.items()}
    else:
        normalized_ref = ref

    min_value_spe = min(spe_pos_dict.values())
    max_value_spe = max(spe_pos_dict.values())
    # Min-Max normalize the spe_pos_dict
    if len(spe_pos_dict.keys()) > 1:
        normalized_spe = {
                key: (value - min_value_spe) / (max_value_spe - min_value_spe) for key, value in spe_pos_dict.items()
                }
    else:
        normalized_spe = spe_pos_dict
    data_list = [
        {'Wavelength': key, 'Intensity': value, 'Source': 'spe'} for key, value in normalized_spe.items()
    ] + [
        {'Wavelength': key, 'Intensity': value, 'Source': 'reference'} for key, value in normalized_ref.items()
    ]

    # Create a DataFrame from the list of dictionaries
    df = pd.DataFrame(data_list)
    feature_matrix = df[['Wavelength', 'Intensity']].to_numpy()

    n_ref = len(ref.keys())
    n_spe = len(spe_pos_dict.keys())
    kmeans = KMeans(n_clusters=n_ref if n_ref > n_spe else n_spe)
    kmeans.fit(feature_matrix)
    labels = kmeans.labels_
    # Extract cluster labels, x values, and y values
    df['Cluster'] = labels
    grouped = df.groupby('Cluster')
    x_spe = np.array([])
    x_reference = np.array([])
    x_distance = np.array([])
    clusters = np.array([])
    # Iterate through each group
    for cluster, group in grouped:
        # Get the unique sources within the group
        unique_sources = group['Source'].unique()
        if 'reference' in unique_sources and 'spe' in unique_sources:
            # Pivot the DataFrame to create the desired structure
            for w_spe in group.loc[group["Source"] == "spe"]["Wavelength"].values:
                x = None
                r = None
                e_min = None
                for w_ref in group.loc[group["Source"] == "reference"]["Wavelength"].values:
                    e = euclidean_distances(w_spe.reshape(-1, 1), w_ref.reshape(-1, 1))[0][0]
                    if (e_min is None) or (e < e_min):
                        x = w_spe
                        r = w_ref
                        e_min = e
                x_spe = np.append(x_spe, x)
                x_reference = np.append(x_reference, r)
                x_distance = np.append(x_distance, e_min)
                clusters = np.append(clusters, cluster)
    sort_indices = np.argsort(x_spe)
    return (x_spe[sort_indices], x_reference[sort_indices], x_distance[sort_indices], df)



import numpy as np
import struct


def readSPA(filename: str, flip=True):
    """
    function to read k-vector and spectrum from a *.SPA file

    Args:
        filename:
            Full path to the file to be read.

    Returns:
        k-vector and spectrum as separate arrays: `(np.array, np.array)`.
    """
    k = None
    spec = None
    with open(filename, 'rb') as f:
        # the data offset is saved at offset 386
        f.seek(386, 0)
        offset = struct.unpack("i", f.read(4))[0]
        # the number of data points is saved at offset 564
        f.seek(564, 0)
        n = struct.unpack("i", f.read(4))[0]
        # the max and min wavenumbers are saved at 576 and 580 respectively
        f.seek(576, 0)
        w_max = struct.unpack("f", f.read(4))[0]
        w_min = struct.unpack("f", f.read(4))[0]
        k = np.linspace(w_min, w_max, n)
        # read the data points
        f.seek(offset, 0)
        spec = np.array([struct.unpack("f", f.read(4))[0] for i in range(n)])
        if flip:
            spec = np.flip(spec)
        # set corrupt counts to zero
        spec[spec > 1e5] = 0
    return k, spec, {}


def read_ngs(file):
    with open(file, "rb") as f:
        # read NextGen string
        s = f.read(10)
        # if the string does not match, abort
        nextgen_string = s.decode("utf-8").lower()
        if nextgen_string != 'ngsnextgen':
            print('Not a readable file !')
            return
        # read DataMatrix string form byte #18
        s = f.seek(18)
        # read length of string as single byte
        s = f.read(1)
        ll = int.from_bytes(s, "big")
        # read the actual string
        s = f.read(ll)
        datamatrix_string = s.decode("utf-8").lower()
        # if the string does not match, abort
        if datamatrix_string != 'datamatrix':
            print('Not a readable file !')
            return
        # read filename at byte #38
        s = f.seek(38)
        # read length of string as single byte
        s = f.read(1)
        ll = int.from_bytes(s, "big")
        # read the actual string
        s = f.read(ll)
        file_name = s.decode("utf-8")
        # Read no. of channels as 32 bit integer, 16 bytes from end of filename
        s = f.seek(16, 1)
        s = f.read(4)
        n = struct.unpack('i', s)[0]
        print(f'Reading Labspec .ngs file {file_name} with {n} channels.')
        # Read data block, starting 8 bytes from end of channel num. Each y count is a 32 bit integer.
        y = read_4byte_datablock(f, n, 8)
        # Read parameter block. Before, there's a rather complicated sequence of skipping obsolete parameters...
        f.seek(2, 1)
        s = f.read(1)
        ll = int.from_bytes(s, "big")
        # Skip bytes as long as they are zeros
        if ll == 0:
            while ll == 0:
                s = f.read(1)
                ll = int.from_bytes(s, "big")
            f.seek(1, 1)
            s = f.read(1)
            ll = int.from_bytes(s, "big")

        f.seek(ll, 1)
        s = f.read(1)
        ll = int.from_bytes(s, "big")

        f.seek(ll, 1)
        f.seek(2, 1)
        s = f.read(1)
        ll = int.from_bytes(s, "big")

        f.seek(ll, 1)
        s = f.read(1)
        ll = int.from_bytes(s, "big")

        f.seek(ll, 1)
        f.seek(16, 1)
        # Finally, read the start and end wavenumbers (start_x)
        s = f.read(4)
        start_x = struct.unpack('f', s)[0]
        s = f.read(4)
        end_x = struct.unpack('f', s)[0]
        # Check whether end_x is equal to no. of channels
        if end_x == n:
            x = read_4byte_datablock(f, n, 20)
        else:
            # Construct x axis
            x = np.linspace(start_x, end_x, n)
        meta = read_ngs_meta(f)
    return x, y, meta


def read_4byte_datablock(f, length, skip=8):
    s = f.seek(skip, 1)
    y = np.zeros(length)
    for ii in range(length):
        s = f.read(4)
        y[ii] = struct.unpack('f', s)[0]
    return y


def read_bytestring(f):
    # read length of string as single byte
    s = f.read(1)
    ll = int.from_bytes(s, "big")
    # read the actual string
    s = f.read(ll)
    return s.decode('iso-8859-1')


def read_ngs_meta(f):
    position = f.tell()+1
    f.seek(position)
    abl = ''
    while abl != 'Table':
        s = f.read(1)
        ll = int.from_bytes(s, "big")
        if ll == 5:
            f.seek(-1, 1)
            abl = read_bytestring(f)
        else:
            position += 1
            f.seek(position)
    abl = ''
    position = f.tell()
    f.seek(position)
    while abl != 'Table':
        s = f.read(1)
        ll = int.from_bytes(s, "big")
        if ll == 5:
            f.seek(-1, 1)
            abl = read_bytestring(f)
            position = f.tell()
        else:
            position += 1
            f.seek(position)
    abl = read_bytestring(f)
    # Make dictionary
    meta = {}
    if abl.upper() == 'ACQ':
        abl = read_bytestring(f)
        f.seek(4, 1)
        # read no. of params
        s = f.read(1)
        ll = int.from_bytes(s, "big")
        f.seek(1, 1)
        par_names = []
        # Read parameter names
        for ii in range(ll):
            abl = read_bytestring(f)
            par_names.append(abl)
        # Skip some stuff
        f.seek(2, 1)
        abl = read_bytestring(f)
        abl = read_bytestring(f)
        f.seek(10, 1)
        s = f.read(1)
        # read no. of values
        ll = int.from_bytes(s, "big")
        f.seek(1, 1)
        # Read parameter values
        par_values = []
        for ii in range(ll):
            abl = read_bytestring(f)
            par_values.append(abl)
        f.seek(2, 1)
        # Read parameter units
        par_units = []
        for ii in range(ll):
            abl = read_bytestring(f)
            # add unit to dict key and exchange for new key
            par_units.append(abl)
        # add units to par names
        par = [name + f' [{unit}]' for name, unit in zip(par_names, par_units)]
        meta = dict(zip(par, par_values))
        return meta



import logging

from pydantic import validate_call

from ramanchada2.misc.spectrum_deco import add_spectrum_method

from ..spectrum import Spectrum

logger = logging.getLogger(__name__)


@add_spectrum_method
@validate_call(config=dict(arbitrary_types_allowed=True))
def fit_peak_positions(spe: Spectrum, /, *,
                       mov_min=40,
                       center_err_threshold=.5,
                       find_peaks_kw={},
                       fit_peaks_kw={},
                       ):
    ss = spe.subtract_moving_minimum(mov_min)  # type: ignore
    find_kw = dict(sharpening=None)
    find_kw.update(find_peaks_kw)
    cand = ss.find_peak_multipeak(**find_kw)

    fit_kw = dict(profile='Gaussian')
    fit_kw.update(fit_peaks_kw)
    fit_res = spe.fit_peak_multimodel(candidates=cand, **fit_kw)  # type: ignore

    pos, amp = fit_res.center_amplitude(threshold=center_err_threshold)

    return dict(zip(pos, amp))



import datetime
import json
from typing import Any, Dict, List, Union

import numpy as np
import numpy.typing as npt
from pydantic import Field, StrictBool, StrictInt, StrictStr, field_validator

from ..pydantic_base_model import PydBaseModel, PydRootModel

SpeMetadataFieldTyping = Union[
    npt.NDArray,
    StrictBool,
    StrictInt, float,
    datetime.datetime,
    List[Any], Dict[str, Any],
    StrictStr]


class SpeMetadataFieldModel(PydRootModel):
    root: SpeMetadataFieldTyping = Field(union_mode='left_to_right')

    @field_validator('root', mode='before')
    def pre_validate(cls, val):
        if isinstance(val, np.ndarray):
            return val
        if isinstance(val, str):
            if val.startswith('ramanchada2_model@'):
                # The format is:
                # ramanchada2_model@ModelName#<DATA>
                pos_at = val.index('@')
                pos_hash = val.index('#')
                model_name = val[pos_at+1:pos_hash]
                from ramanchada2.misc import types
                model = getattr(types, model_name)
                return model.model_validate(val[pos_hash+1:])
            if (val.startswith('[') and val.endswith(']') or
               val.startswith('{') and val.endswith('}')):
                return json.loads(val.replace("'", '"').replace(r'b"', '"'))
        return val

    def serialize(self):
        if isinstance(self.root, list) or isinstance(self.root, dict):
            return json.dumps(self.root)
        if isinstance(self.root, PydBaseModel):
            return f'ramanchada2_model@{type(self.root).__name__}#' + self.json()
        if isinstance(self.root, datetime.datetime):
            return self.root.isoformat()
        if isinstance(self.root, PydBaseModel):
            return self.root.serialize()
        return self.root


class SpeMetadataModel(PydRootModel):
    root: Dict[str, SpeMetadataFieldModel]

    def __str__(self):
        return str(self.serialize())

    def serialize(self):
        return {k: v.serialize() for k, v in sorted(self.root.items())}

    def __getitem__(self, key: str) -> SpeMetadataFieldTyping:
        return self.root[key].root

    def _update(self, val: Dict):
        self.root.update(self.model_validate(val).root)

    def _del_key(self, key: str):
        del self.root[key]

    def _flush(self):
        self.root = {}

    def get_all_keys(self) -> list[str]:
        """
        Returns a list of all keys in the metadata model.
        """
        return list(self.root.keys())



import matplotlib.pyplot as plt
import numpy as np
import pickle
import ramanchada2.misc.constants as rc2const
from ..misc import utils as rc2utils
from ..spectrum import Spectrum
from matplotlib.axes import Axes
from ramanchada2.misc.plottable import Plottable
from scipy import interpolate
import logging
import json
import os
from typing import Tuple, Optional
from pydantic import BaseModel, ValidationError

logger = logging.getLogger(__name__)


class ProcessingModel:

    def __init__(self):
        pass


class CalibrationComponent(Plottable):

    def __init__(self, laser_wl, spe, spe_units, ref, ref_units, sample=None):
        super(CalibrationComponent, self).__init__()
        self.laser_wl = laser_wl
        self.spe = spe
        self.spe_units = spe_units
        self.ref = ref
        self.ref_units = ref_units
        self.name = "not estimated"
        self.model = None
        self.model_units = None
        self.peaks = None
        self.sample = sample
        self.enabled = True

    def set_model(self, model, model_units, peaks, name=None):
        self.model = model
        self.model_units = model_units
        self.peaks = peaks
        self.name = "calibration component" if name is None else name

    def __str__(self):
        return (f"{self.name} spe ({self.spe_units}) reference ({self.ref_units}) "
                f"model ({self.model_units}) {self.model}")

    def convert_units(self, old_spe, spe_unit="cm-1", newspe_unit="nm", laser_wl=None):
        if laser_wl is None:
            laser_wl = self.laser_wl
        logger.debug("convert laser_wl {} {} --> {}".format(laser_wl, spe_unit, newspe_unit))
        if spe_unit != newspe_unit:
            new_spe = old_spe.__copy__()
            if spe_unit == "nm":
                new_spe = old_spe.abs_nm_to_shift_cm_1_filter(laser_wave_length_nm=laser_wl)
            elif spe_unit == "cm-1":
                new_spe = old_spe.shift_cm_1_to_abs_nm_filter(laser_wave_length_nm=laser_wl)
            else:
                raise Exception("Unsupported conversion {} to {}", spe_unit, newspe_unit)
        else:
            new_spe = old_spe.__copy__()
        #    new_spe = old_spe.__copy__()
        return new_spe

    def process(self, old_spe: Spectrum, spe_units="cm-1", convert_back=False):
        raise NotImplementedError(self)

    def derive_model(self, find_kw={}, fit_peaks_kw={}, should_fit=False, name=None):
        raise NotImplementedError(self)

    def plot(self, ax=None, label=' ', **kwargs) -> Axes:
        if ax is None:
            fig, ax = plt.subplots(3, 1, figsize=(12, 4))
        self._plot(ax[0], label=label, **kwargs)
        ax.legend()
        return ax

    def _plot(self, ax, **kwargs):
        pass


class XCalibrationComponent(CalibrationComponent):
    def __init__(self, laser_wl, spe, spe_units, ref, ref_units, sample="Neon"):
        super(XCalibrationComponent, self).__init__(laser_wl, spe, spe_units, ref, ref_units, sample)
        self.spe_pos_dict = None

    def process(self, old_spe: Spectrum, spe_units="cm-1", convert_back=False):
        logger.debug("convert spe_units {} --> model units {}".format(spe_units, self.model_units))
        new_spe = self.convert_units(old_spe, spe_units, self.model_units)
        logger.debug("process", self)
        if self.model is None:
            return new_spe
        elif self.enabled:
            if isinstance(self.model, interpolate.RBFInterpolator):
                new_spe.x = self.model(new_spe.x.reshape(-1, 1))
            elif isinstance(self.model, float):
                new_spe.x = new_spe.x + self.model
        else:
            return new_spe
        # convert back
        if convert_back:
            # print("convert back", spe_units)
            return self.convert_units(new_spe, self.model_units, spe_units)
        else:
            return new_spe

    def _plot(self, ax, **kwargs):
        # self.spe.plot(ax=ax[0].twinx(), label=self.spe_units)

        ax.stem(self.spe_pos_dict.keys(), self.spe_pos_dict.values(), linefmt='b-', basefmt=' ',
                label="{} peaks".format(self.sample))
        ax.twinx().stem(self.ref.keys(), self.ref.values(), linefmt='r-', basefmt=' ',
                        label="Reference {}".format(self.sample))
        ax.set_xlabel("{}".format(self.ref_units))
        ax.legend()
        # fig, ax = plt.subplots(1, 1, figsize=(3, 3))
        # ax.scatter(x_spe, x_reference, marker='o')
        # ax.set_xlabel("spectrum x ({})".format(self.ref_units))
        # ax.set_ylabel("reference x ({})".format(self.ref_units))

        # x_plot = np.linspace(min(x_spe), max(x_spe), 20)
        # y_plot = interp(x_plot.reshape(-1, 1))
        # ax.scatter(x_plot, y_plot, marker='.', label=kwargs["kernel"])

    def _plot_peaks(self, ax, **kwargs):
        # self.model.peaks
        pass
        # fig, ax = plt.subplots(3,1,figsize=(12,4))
        # spe.plot(ax=ax[0].twinx(),label=spe_units)
        # spe_to_process.plot(ax=ax[1],label=ref_units)

    def derive_model(self, find_kw={}, fit_peaks_kw={}, should_fit=False, name=None):

        # convert to ref_units
        logger.debug("[{}]: convert spe_units {} to ref_units {}".format(self.name, self.spe_units, self.ref_units))
        spe_to_process = self.convert_units(self.spe, self.spe_units, self.ref_units)
        logger.debug("max x", max(spe_to_process.x), self.ref_units)
        # spe_to_process.plot(ax=ax[1], label=self.ref_units)

        # if should_fit:
        #     spe_pos_dict = spe_to_process.fit_peak_positions(center_err_threshold=1,
        #                    find_peaks_kw=find_kw, fit_peaks_kw=fit_peaks_kw)  # type: ignore
        # else:
        #     find_kw = dict(sharpening=None)
        #     spe_pos_dict = spe_to_process.find_peak_multipeak(**find_kw).get_pos_ampl_dict()  # type: ignore
        # prominence = prominence, wlen=wlen, width=width
        find_kw = dict(sharpening=None)
        if should_fit:
            self.spe_pos_dict = spe_to_process.fit_peak_positions(
                    center_err_threshold=10, find_peaks_kw=find_kw, fit_peaks_kw=fit_peaks_kw)  # type: ignore
            # fit_res = spe_to_process.fit_peak_multimodel(candidates=cand, **fit_peaks_kw)
            # pos, amp = fit_res.center_amplitude(threshold=1)
            # spe_pos_dict = dict(zip(pos, amp))
        else:
            # prominence = prominence, wlen=wlen, width=width
            print("find_peak_multipeak")
            cand = spe_to_process.find_peak_multipeak(**find_kw)
            self.spe_pos_dict = cand.get_pos_ampl_dict()
        x_spe, x_reference, x_distance, df = rc2utils.match_peaks(self.spe_pos_dict, self.ref)
        sum_of_differences = np.sum(np.abs(x_spe - x_reference)) / len(x_spe)
        logger.debug("sum_of_differences original {} {}".format(sum_of_differences, self.ref_units))
        if len(x_reference) == 1:
            _offset = (x_reference[0] - x_spe[0])
            logger.debug("ref", x_reference[0], "sample", x_spe[0], "offset", _offset, self.ref_units)
            self.set_model(_offset, self.ref_units, df, name)
        else:
            try:
                kwargs = {"kernel": "thin_plate_spline"}
                interp = interpolate.RBFInterpolator(x_spe.reshape(-1, 1), x_reference, **kwargs)
                self.set_model(interp, self.ref_units, df, name)
            except Exception as err:
                raise err


class LazerZeroingComponent(CalibrationComponent):
    def __init__(self, laser_wl, spe, spe_units="nm", ref={520.45: 1}, ref_units="cm-1", sample="Silicon"):
        super(LazerZeroingComponent, self).__init__(laser_wl, spe, spe_units, ref, ref_units, sample)
        self.profile = "Pearson4"

    def derive_model(self, find_kw={}, fit_peaks_kw={}, should_fit=True, name=None):
        find_kw = dict(sharpening=None)
        cand = self.spe.find_peak_multipeak(**find_kw)
        logger.debug(self.name, cand)
        # init_guess = self.spe.fit_peak_multimodel(profile='Pearson4', candidates=cand, no_fit=False)
        fit_res = self.spe.fit_peak_multimodel(profile=self.profile, candidates=cand, **fit_peaks_kw)
        df = fit_res.to_dataframe_peaks()
        df = df.sort_values(by='height', ascending=False)
        if df.empty:
            raise Exception("No peaks found")
        else:
            if "position" in df.columns:
                zero_peak_nm = df.iloc[0]["position"]
            elif "center" in df.columns:
                zero_peak_nm = df.iloc[0]["center"]
            # print(self.name, "peak", zero_peak_nm)
            # https://www.elodiz.com/calibration-and-validation-of-raman-instruments/
            self.set_model(zero_peak_nm, "nm", df, "Lazer zeroing using {} nm".format(zero_peak_nm))
            logger.info(self.name, "peak", zero_peak_nm)
        # laser_wl should be calculated  based on the peak position and set instead of the nominal

    def zero_nm_to_shift_cm_1(self, wl, zero_pos_nm, zero_ref_cm_1=520.45):
        return 1e7*(1/zero_pos_nm - 1/wl) + zero_ref_cm_1

    # we do not do shift (as initially implemented)
    # just convert the spectrum nm->cm-1 using the Si measured peak in nm and reference in cm-1
    # https://www.elodiz.com/calibration-and-validation-of-raman-instruments/
    def process(self, old_spe: Spectrum, spe_units="nm", convert_back=False):
        wl_si_ref = list(self.ref.keys())[0]
        logger.debug(self.name, "process", self.model, wl_si_ref)
        new_x = self.zero_nm_to_shift_cm_1(old_spe.x, self.model, wl_si_ref)
        new_spe = Spectrum(x=new_x, y=old_spe.y, metadata=old_spe.meta)
        # new_spe = old_spe.lazer_zero_nm_to_shift_cm_1(self.model, wl_si_ref)
        # print("old si", old_spe.x)
        # print("new si", new_spe.x)
        return new_spe

    def _plot(self, ax, **kwargs):
        # spe_sil.plot(label="{} original".format(si_tag),ax=ax)
        # spe_sil_calib.plot(ax = ax,label="{} laser zeroed".format(si_tag),fmt=":")
        # ax.set_xlim(520.45-50,520.45+50)
        # ax.set_xlabel("cm-1")
        pass


class YCalibrationCertificate(BaseModel, Plottable):
    """
    Class for intensity calibration certificates

    Usage:

        1. Use for specific SRM
        >>> cert = YCalibrationCertificate(
        ...             id="NIST785_SRM2241",
        ...             description="optical glass",
        ...             url="https://tsapps.nist.gov/srmext/certificates/2241.pdf",
        ...             wavelength=785,
        ...             params="A0 = 9.71937e-02, A1 = 2.28325e-04, A2 = -5.86762e-08, A3 = 2.16023e-10, A4 = -9.77171e-14, A5 = 1.15596e-17",
        ...             equation="A0 + A1 * x + A2 * x**2 + A3 * x**3 + A4 * x**4 + A5 * x**5",
        ...             temperature_c=(20, 25),
        ...             raman_shift=(200, 3500)
        ...         )
        ...
        >>> cert.plot()

    """  # noqa: E501

    id: str
    description: Optional[str]
    url: Optional[str]
    wavelength: int
    params: str
    equation: str
    temperature_c: Optional[Tuple[int, int]]
    raman_shift: Optional[Tuple[int, int]]

    @property
    def response_function(self):
        local_vars = {}
        for param in self.params.split(','):
            key, value = param.split('=')
            key = key.strip()
            value = value.strip()
            local_vars[key] = eval(value)

        def evaluate_expression(x_value):
            local_vars['x'] = x_value
            return eval(self.equation, {"np": np}, local_vars)

        return evaluate_expression

    def Y(self,  x_value):
        return self.response_function(x_value)

    def _plot(self, ax, **kwargs):
        if self.raman_shift is None:
            x = np.linspace(100, 4000)
        else:
            x = np.linspace(self.raman_shift[0], self.raman_shift[1])
        kwargs.pop('label', None)
        ax.plot(x, self.Y(x), label="{} ({}nm)".format(self.id, self.wavelength), **kwargs)
        ax.set_xlabel('Raman shift cm-1')
        ax.set_ylabel('Intensity')

    @staticmethod
    def load(wavelength=785, key="NIST785_SRM2241"):
        return CertificatesDict().get(wavelength, key)


class CertificatesDict:
    """
    Class for loading y calibration certificates

    Usage:
       Load single certificate
       >>> cert = CertificatesDict.load(wavelength="785", key="NIST785_SRM2241")
       >>> cert.plot()

       Load all certificates for wavelength. Iterate :

        >>> certificates = CertificatesDict()
        ... plt.figure()
        ... ax=None
        ... certs = certificates.get_certificates(wavelength=532)
        ... ax = certs[cert].plot(ax=ax)
        >>> plt.show()
    """
    def __init__(self):
        self.load_certificates(os.path.join(os.path.dirname(__file__), "config_certs.json"))

    def load_certificates(self, file_path):

        with open(file_path, 'r') as f:
            certificates_data = json.load(f)
            certificates = {}
            self.laser_wl = []
            for wavelength, certificates_dict in certificates_data.items():
                certificates[wavelength] = {}
                self.laser_wl.append(wavelength)
                for certificate_id, certificate_data in certificates_dict.items():
                    certificate_data["wavelength"] = int(wavelength)
                    certificate_data["id"] = certificate_id
                    try:
                        certificate = YCalibrationCertificate(**certificate_data)
                        certificates[wavelength][certificate_id] = certificate
                    except ValidationError as e:
                        print(f"Validation error for certificate {certificate_id}: {e}")
            self.config_certs = certificates

    def get_laser_wl(self):
        return self.laser_wl

    def get_certificates(self, wavelength=785):
        return self.config_certs[str(wavelength)]

    def get(self, wavelength=532, key="NIST532_SRM2242a"):
        return self.config_certs[str(wavelength)][key]

    @staticmethod
    def load(wavelength=785, key="NIST785_SRM2241"):
        return CertificatesDict().get(wavelength, key)


class YCalibrationComponent(CalibrationComponent):
    """
    Class for intensity calibration. Uses response functions loaded in ResponseFunctionEvaluator. Functions are defined
    in json file.

    Usage:

        >>> laser_wl = 785
        >>> ycert = YCalibrationCertificate.load(wavelength=785, key="SRM2241")
        >>> ycal = YCalibrationComponent(laser_wl, reference_spe_xcalibrated=spe_srm,certificate=ycert)
        >>> fig, ax = plt.subplots(1, 1, figsize=(15,4))
        >>> spe_srm.plot(ax=ax)
        >>> spe_to_correct.plot(ax=ax)
        >>> spe_ycalibrated = ycal.process(spe_to_correct)
        >>> spe_ycalibrated.plot(label="y-calibrated",color="green",ax=ax.twinx())
    """

    def __init__(self, laser_wl, reference_spe_xcalibrated, certificate: YCalibrationCertificate):
        super(YCalibrationComponent, self).__init__(laser_wl, spe=reference_spe_xcalibrated, spe_units=None,
                                                    ref=certificate, ref_units=None)
        self.laser_wl = laser_wl
        self.spe = reference_spe_xcalibrated
        self.ref = certificate
        self.name = "Y calibration"
        self.model = self.spe.spe_distribution(trim_range=certificate.raman_shift)
        self.model_units = "cm-1"

    def derive_model(self, find_kw={}, fit_peaks_kw={}, should_fit=True, name=None):
        # measured reference spectrum as distribution, so we can resample
        self.model = self.spe.spe_distribution(trim_range=self.ref.raman_shift)

    def safe_divide(self, spe_to_correct, spe_reference_resampled):
        numerator = spe_to_correct.y
        # numerator_noise = spe_to_correct.y_noise

        scaling_denominator = spe_reference_resampled.y / self.ref.Y(spe_reference_resampled.x)
        print(np.median(scaling_denominator), np.mean(scaling_denominator), np.std(scaling_denominator))

        # denominator_noise = spe_reference_resampled.y_noise
        denominator = spe_reference_resampled.y
        # Create a mask for dividing only where value is above noise !
        # mask = (abs(scaling_denominator) > 0) & (kind_of_snr > 0.9)
        # mask =  (abs(denominator) > abs(denominator_noise)) &
        mask = (abs(scaling_denominator) > 0) & (numerator > 0) & (denominator > 0)
        # & (abs(numerator) > numerator_noise) & (abs(scaling_denominator) > 0)
        # & (abs(denominator-numerator) > min(denominator_noise,numerator_noise))
        result = np.zeros_like(numerator)
        # Perform division where mask is true
        result[mask] = numerator[mask] / scaling_denominator[mask]
        return result

    def safe_mask(self, spe_to_correct, spe_reference_resampled):
        ref_noise = spe_reference_resampled.y_noise
        return (spe_reference_resampled.y >= 0) & (abs(spe_reference_resampled.y) > ref_noise)

    def safe_factor(self, spe_to_correct, spe_reference_resampled):
        numerator = spe_to_correct.y
        # numerator_noise = spe_to_correct.y_noise

        Y = self.ref.Y(spe_reference_resampled.x)
        mask = self.safe_mask(spe_to_correct, spe_reference_resampled)
        if mask is None:
            scaling_factor = Y / spe_reference_resampled.y
        else:
            scaling_factor = np.zeros_like(spe_reference_resampled.y)
            scaling_factor[mask] = Y[mask] / spe_reference_resampled.y[mask]

        result = numerator * scaling_factor
        return result

    def process(self, old_spe: Spectrum, spe_units="nm", convert_back=False):
        # resample using probability density function
        _tmp = self.model.pdf(old_spe.x)
        _tmp = _tmp * max(self.spe.y) / max(_tmp)  # pdf sampling is normalized to area unity, scaling back
        spe_reference_resampled = Spectrum(old_spe.x, _tmp)
        # new_spe = Spectrum(old_spe.x,self.safe_divide(old_spe,spe_reference_resampled))
        new_spe = Spectrum(old_spe.x, self.safe_factor(old_spe, spe_reference_resampled))
        return new_spe

    def _plot(self, ax, **kwargs):
        if self.ref is not None:
            self.ref.plot(ax, **kwargs)


class CalibrationModel(ProcessingModel, Plottable):
    """
    A class representing a calibration model for Raman spectrum.
    """

    def __init__(self, laser_wl: int):
        """
        Initializes a CalibrationModel instance.

        Args:
            laser_wl:
                The wavelength of the laser used for calibration.

        Example:
        ```python
        # Create an instance of CalibrationModel
        calmodel = CalibrationModel(laser_wl=785)
        calmodel.derive_model_x(
            spe_neon,
            spe_neon_units="cm-1",
            ref_neon=None,
            ref_neon_units="nm",
            spe_sil=None,
            spe_sil_units="cm-1",
            ref_sil=None,
            ref_sil_units="cm-1"
            )
        # Store
        calmodel.save(modelfile)
        # Load
        calmodel = CalibrationModel.from_file(modelfile)
        # Apply to new spectrum
        calmodel.apply_calibration_x(
            spe_to_calibrate,
            spe_units="cm-1"
            )
        ```
        """
        super(ProcessingModel, self).__init__()
        super(Plottable, self).__init__()
        self.set_laser_wavelength(laser_wl)
        self.neon_wl = {
            785: rc2const.neon_wl_785_nist_dict,
            633: rc2const.neon_wl_633_nist_dict,
            532: rc2const.neon_wl_532_nist_dict
        }
        self.prominence_coeff = 10

    def peaks(self, spe, profile='Gaussian', wlen=300, width=1):
        """
        Finds and fits peaks in the spectrum spe.
        """
        cand = spe.find_peak_multipeak(prominence=spe.y_noise*self.prominence_coeff, wlen=wlen, width=width)
        init_guess = spe.fit_peak_multimodel(profile=profile, candidates=cand, no_fit=True)
        fit_res = spe.fit_peak_multimodel(profile=profile, candidates=cand)
        return cand, init_guess, fit_res

    def set_laser_wavelength(self, laser_wl):
        """
        Sets the wavelength of the laser used for calibration.
        """
        self.clear()
        self.laser_wl = laser_wl

    def clear(self):
        """
        Clears the calibration model.
        """
        self.laser_wl = None
        self.components = []

    def save(self, filename):
        """
        Saves the calibration model to a file.
        """
        with open(filename, "wb") as file:
            pickle.dump(self, file)

    @staticmethod
    def from_file(filename):
        """
        Loads a calibration model from a file.
        """
        with open(filename, "rb") as file:
            return pickle.load(file)

    def derive_model_x(self, spe_neon, spe_neon_units="cm-1", ref_neon=None, ref_neon_units="nm", spe_sil=None,
                       spe_sil_units="cm-1", ref_sil=None, ref_sil_units="cm-1", find_kw={}, fit_kw={}):
        """
        Derives x-calibration models using Neon and Silicon spectra.
        """
        model_neon = self.derive_model_curve(
                spe_neon, self.neon_wl[self.laser_wl], spe_units=spe_neon_units, ref_units=ref_neon_units, find_kw={},
                fit_peaks_kw={}, should_fit=False, name="Neon calibration")
        spe_sil_ne_calib = model_neon.process(spe_sil, spe_units=spe_sil_units, convert_back=False)
        find_kw = {"prominence": spe_sil_ne_calib.y_noise * 10, "wlen": 200, "width":  1}
        model_si = self.derive_model_zero(
                spe_sil_ne_calib, ref={520.45: 1}, spe_units="nm", ref_units=ref_sil_units, find_kw=find_kw,
                fit_peaks_kw={}, should_fit=True, name="Si laser zeroing")
        return (model_neon, model_si)

    def derive_model_curve(self, spe, ref, spe_units="cm-1", ref_units="nm", find_kw={}, fit_peaks_kw={},
                           should_fit=False, name="X calibration"):
        calibration_x = XCalibrationComponent(self.laser_wl, spe, spe_units, ref, ref_units)
        calibration_x.derive_model(find_kw=find_kw, fit_peaks_kw=fit_peaks_kw, should_fit=should_fit, name=name)
        self.components.append(calibration_x)
        return calibration_x

    def derive_model_zero(self, spe, ref, spe_units="nm", ref_units="cm-1", find_kw={}, fit_peaks_kw={},
                          should_fit=False, name="X Shift", profile="Gaussian"):
        calibration_shift = LazerZeroingComponent(self.laser_wl, spe, spe_units, ref, ref_units)
        calibration_shift.profile = profile
        calibration_shift.derive_model(find_kw=find_kw, fit_peaks_kw=fit_peaks_kw, should_fit=should_fit, name=name)
        self.components.append(calibration_shift)
        return calibration_shift

    def apply_calibration_x(self, old_spe: Spectrum, spe_units="cm-1"):
        # neon calibration converts to nm
        # silicon calibration takes nm and converts back to cm-1 using laser zeroing
        new_spe = old_spe
        model_units = spe_units
        for model in self.components:
            # TODO: tbd find out if to convert units
            if model.enabled:
                new_spe = model.process(new_spe, model_units, convert_back=False)
                model_units = model.model_units
        return new_spe

    def plot(self, ax=None, label=' ', **kwargs) -> Axes:
        if ax is None:
            fig, ax = plt.subplots(1, 1, figsize=(12, 4))
        self._plot(ax, **kwargs)
        return ax

    def _plot(self, ax, **kwargs):
        for index, model in enumerate(self.components):
            model._plot(ax, **kwargs)
            break



#!/usr/bin/env python3

from __future__ import annotations
import numpy as np

from uncertainties import unumpy

from ..spectral_peak import SpectralPeak


class DeltasPeak(SpectralPeak):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if {'a', 'x0'} - kwargs.keys():
            raise ValueError("'a', 'x0' arguments required")
        self.a = kwargs['a']
        self.x0 = kwargs['x0']

    def __call__(self, x: unumpy.uarray):
        ret = np.zeros_like(x)
        ret[x == self.x0] = self.a
        return ret

    @property
    def delta(self):
        return self.x0, self.a

    @property
    def pos_amp_fwhm(self):
        return self.x0, self.a, 0

    @property
    def limit_3sigma(self):
        return self.x0-1, self.x0+1



from typing import Dict, List, Literal, Tuple

import numpy as np
from lmfit import Model, Parameters
from lmfit.models import GaussianModel, VoigtModel
from pydantic import BaseModel, Field, validate_call


@validate_call(config=dict(arbitrary_types_allowed=True))
def model_from_lines(names: List[str],
                     positions: List[float],
                     intensities: Dict[str, List[float]],
                     model: Literal['gaussian', 'voigt'] = 'gaussian'
                     ) -> Tuple[Model, Parameters]:

    if model == 'gaussian':
        lm_model = GaussianModel
    elif model == 'voigt':
        lm_model = VoigtModel
    else:
        raise ValueError(f'model {model} not known')
    mod = np.sum([
        lm_model(prefix=f'{spe_type}_{name}_', name=f'{spe_type}_{name}')
        for spe_type in intensities
        for name in names
        ])

    params = Parameters()
    params.add('pedestal', 0, min=0)
    params.add('sigma', 2, min=0)
    params.add('x0', 0)
    params.add('x1', 1, min=0)

    for spe_type, spe_int in intensities.items():
        spe_prefix = f'{spe_type}_'
        params.add(spe_prefix+'amplitude', 1, min=0)
        for name, pos, line_int in zip(names, positions, spe_int):
            line_prefix = f'{spe_prefix}{name}_'
            params.add(line_prefix+'amplitude',
                       expr=f'({line_int}*{spe_prefix}amplitude)+pedestal')
            params.add(line_prefix+'center', expr=f'(({pos}+x0)*x1)')
            params.add(line_prefix+'sigma', expr='sigma')

    return mod, params


class PydPeakModel(BaseModel):
    model: Literal['gaussian', 'voigt'] = Field('voigt')
    position: float
    inensity: float = Field(1, gt=0)
    sigma: float = Field(1, gt=0)
    name: str = Field('')


@validate_call(config=dict(arbitrary_types_allowed=True))
def model_from_list(peaks_list: List[PydPeakModel]
                    ) -> Tuple[Model, Parameters]:
    params = Parameters()
    params.add('amplitude', 1, min=0)
    params.add('sigma', 1, min=0)
    params.add('x0', 0)
    params.add('x1', 1, min=0)
    params.add('x2', 0, min=-1e-3, max=1e-3)
    params.add('x3', 0, min=-1e-5, max=1e-5)

    peaks = list()
    for ii, peak in enumerate(peaks_list):
        if peak.model == 'gaussian':
            lm_model = GaussianModel
        elif peak.model == 'voigt':
            lm_model = VoigtModel
        else:
            raise ValueError(f'model {peak.model} not known')
        prefix = f'{peak.name}_' if peak.name else f'_{ii}_'
        name = f'{peak.name}' if peak.name else f'_{ii}'
        peaks.append(lm_model(prefix=prefix, name=name))

        params.add(prefix+'amplitude', expr=f'({peak.inensity}*amplitude)')
        params.add(prefix+'center',
                   expr=f'{peak.position}**3*x3 + {peak.position}**2*x2 + {peak.position}*x1 + x0')
        params.add(prefix+'sigma', expr=f'{peak.sigma}*sigma')
    mod = np.sum(peaks)

    return mod, params



from typing import Union

from pydantic import PositiveInt, validate_call
from sklearn.mixture import BayesianGaussianMixture

from ramanchada2.misc.spectrum_deco import add_spectrum_method

from ..spectrum import Spectrum


@add_spectrum_method
@validate_call(config=dict(arbitrary_types_allowed=True))
def bayesian_gaussian_mixture(spe: Spectrum, /,
                              n_samples: PositiveInt = 5000,
                              n_components: PositiveInt = 50,
                              max_iter: PositiveInt = 100,
                              moving_minimum_window: Union[PositiveInt, None] = None,
                              random_state=None,
                              trim_range=None,
                              ):
    if moving_minimum_window is not None:
        spe = spe.subtract_moving_minimum(moving_minimum_window)  # type: ignore
    samp = spe.gen_samples(size=n_samples, trim_range=trim_range)
    X = [[i] for i in samp]
    bgm = BayesianGaussianMixture(n_components=n_components,
                                  random_state=random_state,
                                  max_iter=max_iter
                                  ).fit(X)
    return bgm
